/*==========================================================================
  PRODUCT_VERSION_STR "15.3.7.24531"
  --------------------------------------------------------------------------
  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.

  All rights reserved. All hardware and software names used are registered
  trade names and/or registered trademarks of the respective manufacturers.

  The user of this computer program acknowledges that the above copyright
  notice, which constitutes the Universal Copyright Convention, will be
  attached at the position in the function of the computer program which the
  author has deemed to sufficiently express the reservation of copyright.
  It is prohibited for customers, users and/or third parties to remove,
  modify or move this copyright notice.
  ==========================================================================*/

#define APMEXPORT
#define APM_WEB_API
#define APM_WEB_API_PERSIST
#include "softwaretelemetry.h"

#if defined(APM_WIN)
# define _CRT_SECURE_NO_WARNINGS
# define _CRT_SECURE_NO_DEPRECATE
# include <windows.h>
#elif defined(APM_UNIX)
# include <unistd.h>
# include <dlfcn.h>
# include <string.h>
# include <sys/time.h>
# include <sys/syscall.h>
#endif

#define SWT_ENTRY_APPLICATION                1
#define SWT_ENTRY_MODULE                     2
#define SWT_ENTRY_FILTER                     3
#define SWT_ENTRY_MODULE_EVENT               4
#define SWT_ENTRY_GET_CONTEXT                5
#define SWT_ENTRY_ATTACH_CONTEXT             6
#define SWT_ENTRY_EVENT                      7
#define SWT_ENTRY_CREATE_CONTEXT             8
#define SWT_ENTRY_RELEASE_CONTEXT            9
#define SWT_ENTRY_SESSION_INFO              10
#define SWT_ENTRY_LIFECHECK_RESPONSE        12
#define SWT_ENTRY_LOST_DATA                 13
#define SWT_ENTRY_APPLICATION_UNREGISTER    14
#define SWT_ENTRY_GET_SYNC_MARK             15
#define SWT_ENTRY_SET_SYNC_MARK             16
#define SWT_ENTRY_SYNC_CONTEXT              17
#define SWT_ENTRY_REPORT                    18
#define SWT_ENTRY_REPORT_VALUE              19
#define SWT_ENTRY_REPORT_FILE_HEADER        20
#define SWT_ENTRY_REPORT_FILE_CONTENT       21
#define SWT_ENTRY_UNREGISTER_MODULE         23
#define SWT_ENTRY_DETACH_THREAD             24
#define SWT_ENTRY_ATTACH_THREAD             25
#define SWT_ENTRY_REPORT_FILE_HEADER_MIME   26
#define SWT_ENTRY_TCP_INIT                  27
#define SWT_ENTRY_REPORT_FORM               28

#if defined(APM_COMPILE_LIBRARY)
# include <version.h>
#else
# define PRODUCT_VERSION_STR "15.3.7.24531"
#endif

#include <string>
#include <sstream>
#include <deque>
#include <vector>
#include <map>
#include <set>
#include <cstdlib>

#if !defined(UNREFERENCED_PARAMETER)
# define UNREFERENCED_PARAMETER(P) (void)(P)
#endif

#if !defined(APM_WEB_API_DEFAULT_BUFFERSIZE)
# define APM_WEB_API_DEFAULT_BUFFERSIZE (10*1024*1024)
#endif

// helper functions

#if defined(WIN32)
# define _apmthread_t DWORD
# define APM_ATOMICTYPE LONG
# define _apm_interlocked_exchange InterlockedExchange

static void _apm_yield()
{
  Sleep(1);
}
#elif defined(MACOSX)
# define _apmthread_t unsigned long
#elif defined(APM_UNIX)
# define _apmthread_t pid_t
#endif

#if defined(APM_UNIX)

# if defined(_X64_)
#  define APM_ATOMICTYPE int
#else
#  define APM_ATOMICTYPE long
# endif

static long _apm_interlocked_exchange(APM_ATOMICTYPE volatile *mem, APM_ATOMICTYPE source)
{
  APM_ATOMICTYPE result;

  __asm__ __volatile__("lock; xchg %0, (%1)"
                       : "=r" (result)
                       : "r" (mem), "0" (source)
                       : "memory");
  return result;
}

static void _apm_yield()
{
  usleep(1);
}

#endif

static _apmthread_t _getThreadId()
{
#if defined(APM_WIN)
  return GetCurrentThreadId();
#elif defined(MACOSX)
  return (_apmthread_t)pthread_self();
#elif defined(APM_UNIX)
  return ((pid_t) syscall (SYS_gettid));
#endif
}

static void toString(UInt64 id, char *buffer)
{
  if (id) {
    char *begin = buffer;
    while (id) {
      *buffer++ = '0' + (char)(id % 10);
      id /= 10;
    }
    *buffer-- = 0;
    while (begin < buffer) {
      char c = *begin;
      *begin = *buffer;
      *buffer = c;
      begin++;
      buffer--;
    }
  }
  else {
    *buffer++ = '0';
    *buffer = 0;
  }
}

static std::string toLowerString(const std::string &src)
{
  std::string result;
  result.reserve(src.length());
  std::string::const_iterator iter;
  for (iter = src.begin(); iter != src.end(); ++iter) {
    char c = *iter;
    if ('A' <= c && c <= 'Z') {
      result.push_back(c - 'A' + 'a');
    }
    else {
      result.push_back(c);
    }
  }
  return result;
}

static bool WriteBuffer(char *&buffer, unsigned int &size, char value)
{
  if (size > 0) {
    *buffer = value;
    buffer++;
    size--;
    return true;
  }
  return false;
}

static bool WriteBuffer(char *&buffer, unsigned int &size, const char *value, size_t length)
{
  if (length <= size) {
    memcpy(buffer, value, length);
    buffer += length;
    size -= (unsigned int)length;
    return true;
  }
  return false;
}

static bool WriteBuffer(char *&buffer, unsigned int &size, const char *value)
{
  return WriteBuffer(buffer, size, value, strlen(value));
}

static bool WriteBuffer(char *&buffer, unsigned int &size, const std::string &value)
{
  return WriteBuffer(buffer, size, value.c_str(), value.length());
}

static bool WriteBufferNumeric(char *&buffer, unsigned int &size, UInt64 number)
{
  char stringbuffer[30];
  toString(number, stringbuffer);
  return WriteBuffer(buffer, size, stringbuffer);
}

static bool WriteBufferStringEscape(char *&buffer, unsigned int &size, const char *value, size_t length)
{
  const char *begin = value;
  const char *end = value;
  bool success = WriteBuffer(buffer, size, '"');
  while (success && length) {
    switch (*end) {
    case '"':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\\"");
      begin = end + 1;
      break;
    case '\\':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\\\");
      begin = end + 1;
      break;
    case '/':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\/");
      begin = end + 1;
      break;
    case '\b':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\b");
      begin = end + 1;
      break;
    case '\f':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\f");
      begin = end + 1;
      break;
    case '\n':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\n");
      begin = end + 1;
      break;
    case '\r':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\r");
      begin = end + 1;
      break;
    case '\t':
      success = WriteBuffer(buffer, size, begin, end - begin) &&
                WriteBuffer(buffer, size, "\\t");
      begin = end + 1;
      break;
    default:
      if ((unsigned char)*begin < 32) {
        success = WriteBuffer(buffer, size, begin, end - begin);
        begin = end + 1;
      }
      break;
    }
    end++;
    length--;
  }
  success = success && WriteBuffer(buffer, size, begin, end - begin) &&
                       WriteBuffer(buffer, size, '"');
  return success;
}

static bool WriteBufferStringEscape(char *&buffer, unsigned int &size, const std::string &value)
{
  return WriteBufferStringEscape(buffer, size, value.c_str(), value.length());
}

#if defined(APM_WIN)
static bool g_qpc_valid = true;
static DWORD g_time_high = 0;
static DWORD g_time_low = 0;
#endif

static UInt64 GetTimestampFrequency()
{
#if defined(APM_WIN)
  LARGE_INTEGER time;
  if (g_qpc_valid) {
    if (QueryPerformanceFrequency(&time)) {
      return time.QuadPart;
    }
    else {
      g_qpc_valid = false;
    }
  }
  return 1000ULL;
#else
  return 1000000ULL;
#endif
}

static UInt64 GetTimestamp()
{
#if defined(APM_WIN)
  LARGE_INTEGER time;
  if (g_qpc_valid) {
    if (QueryPerformanceCounter(&time)) {
      return time.QuadPart;
    }
    else {
      g_qpc_valid = false;
    }
  }
  time.LowPart = GetTickCount(); // use GetTickCount
  if (time.LowPart < g_time_low) { // handle 32-bit rollover every 42 days
    g_time_high++;
  }
  g_time_low = time.LowPart;
  time.HighPart = g_time_high;
  return time.QuadPart;
#else
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (UInt64) tv.tv_sec * 1000000ULL + (UInt64) tv.tv_usec;
#endif
}

template<typename T>
static void WriteItemParamNumeric(std::stringstream &stream, T value)
{
  stream << ',' << value;
}

template <typename T>
static void WriteItemParamArrayNumeric(std::stringstream &stream, std::vector<T> &value)
{
  stream << ',' << '[';
  typename std::vector<T>::iterator iter = value.begin();
  if (iter != value.end()) {
    stream << *iter;
    ++iter;
    for (; iter != value.end(); ++iter) {
      stream << ',' << *iter;
    }
  }
  stream << ']';
}

static void WriteStringEscape(std::stringstream &stream, const char *value)
{
  stream << '"';
  if (value) {
    for (const char *p = value; *p; ++p) {
      switch (*p) {
      case '"':
        stream << "\\\"";
        break;
      case '\\':
        stream << "\\\\";
        break;
      case '/':
        stream << "\\/";
        break;
      case '\b':
        stream << "\\b";
        break;
      case '\f':
        stream << "\\f";
        break;
      case '\n':
        stream << "\\n";
        break;
      case '\r':
        stream << "\\r";
        break;
      case '\t':
        stream << "\\t";
        break;
      default:
        if ((unsigned char)*p >= 32) {
          stream << *p;
        }
        break;
      }
    }
  }
  stream << '"';
}

static void WriteItemParamString(std::stringstream &stream, const char *value)
{
  stream << ',';
  WriteStringEscape(stream, value);
}

static void WriteBase64(std::stringstream &stream, const void *value, size_t length)
{
  static const char bin64digits[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  if (value && length) {
    const unsigned char *p = (const unsigned char *)value;
    while (length > 3) {
      stream << bin64digits[((p[0] & 0xFC) >> 2U)];
      stream << bin64digits[((p[0] & 0x03) << 4U) | ((p[1] & 0xF0) >> 4U)];
      stream << bin64digits[((p[1] & 0x0F) << 2U) | ((p[2] & 0xC0) >> 6U)];
      stream << bin64digits[((p[2] & 0x3F))];
      p += 3;
      length -= 3;
    }
    switch (length) {
    case 1:
      stream << bin64digits[((p[0] & 0xFC) >> 2U)];
      stream << bin64digits[((p[0] & 0x03) << 4U)];
      stream << '=';
      stream << '=';
      break;
    case 2:
      stream << bin64digits[((p[0] & 0xFC) >> 2U)];
      stream << bin64digits[((p[0] & 0x03) << 4U) | ((p[1] & 0xF0) >> 4U)];
      stream << bin64digits[((p[1] & 0x0F) << 2U)];
      stream << '=';
      break;
    }
  }
}

static void WriteItemParamBase64(std::stringstream &stream, const void *value, size_t length)
{
  stream << ',' << '"';
  WriteBase64(stream, value, length);
  stream << '"';
}

enum _JsonTokenType {
  json_object_start=0,
  json_object_end=1,
  json_array_start=2,
  json_array_end=3,
  json_comma_separator=4,
  json_string_token=5,
  json_number_token=6,
  json_float_token=7,
  json_colon_separator=8,
  json_bool_true=9,
  json_bool_false=10,
  json_null=11,
  json_error=12,
  json_end=13
};

#define CASE_JSON_TOKEN(c, token) \
  case c:                         \
    _token = token;               \
    _ptr++;                       \
    break;                        \

class _JSONParser
{
  const char  *_ptr;
  const char  *_end;
  std::string  _value;
  _JsonTokenType _token;
public:
  _JSONParser(const char *buffer, unsigned int size) :
    _ptr(buffer), _end(buffer + size),
    _token(json_error)
  {
  }

  _JsonTokenType getNext()
  {
    while (_ptr != _end && isspace(*_ptr)) {
      _ptr++;
    }
    if (_ptr != _end) {
      switch(*_ptr) {
      CASE_JSON_TOKEN('{', json_object_start);
      CASE_JSON_TOKEN('}', json_object_end);
      CASE_JSON_TOKEN('[', json_array_start);
      CASE_JSON_TOKEN(']', json_array_end);
      CASE_JSON_TOKEN(',', json_comma_separator);
      CASE_JSON_TOKEN(':', json_colon_separator);
      case '"':
      case '\'':
        {
          char delimiter = *_ptr;
          _value.clear();
          _ptr++;
          while (_ptr != _end && *_ptr != delimiter) {
            switch (*_ptr) {
            case '\\':
              _ptr++;
              if (_ptr != _end) {
                switch (*_ptr) {
                case 'b':
                  _value += '\b';
                  break;
                case 'f':
                  _value += '\f';
                  break;
                case 'n':
                  _value += '\n';
                  break;
                case 'r':
                  _value += '\r';
                  break;
                case 't':
                  _value += '\t';
                  break;
                default:
                  _value += *_ptr;
                  break;
                }
              }
              break;
            default:
              _value += *_ptr;
              break;
            }
            _ptr++;
          }
          if (_ptr != _end) {
            _ptr++;
          }
        }
        _token = json_string_token;
        break;
      case '-':
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        _value.clear();
        do {
          _value += *_ptr;
          _ptr++;
        } while (_ptr != _end && '0' <= *_ptr && *_ptr <= '9');
        _token = json_number_token;
        break;
      case 't':
      case 'f':
      case 'n':
        _value.clear();
        do {
          _value += *_ptr;
          _ptr++;
        } while (_ptr != _end && 'a' <= *_ptr && *_ptr <= 'z');
        if (_value == "true") {
          _token = json_bool_true;
        }
        else if (_value == "false") {
          _token = json_bool_false;
        }
        else if (_value == "null") {
          _token = json_null;
        }
        else {
          _token = json_error;
        }
        break;
      default:
        _token = json_error;
        break;
      }
    }
    else {
      _token = json_end;
    }
    return _token;
  }

  _JsonTokenType getToken() const
  {
    return _token;
  }

  const std::string getValue() const
  {
    return _value;
  }

  Int64 getNumber() const
  {
    char *end;
    return strtol(_value.c_str(), &end, 10);
  }

};

class CAPMFilterValue
{
private:
  std::string m_description;
  bool m_refresh;
public:
  CAPMFilterValue(const char *description) :
    m_description(description ? description : ""), m_refresh(true)
  {
  }

  const std::string &getDescription() const
  {
    return m_description;
  }

  void setDescription(const char *description)
  {
    m_description = description ? description : "";
    m_refresh = true;
  }

  bool doRefresh() const
  {
    return m_refresh;
  }

  void setRefresh(bool refresh)
  {
    m_refresh = refresh;
  }

};

typedef std::map<std::string, CAPMFilterValue> CAPMWebFilterValueMap;

class CAPMSession
{
private:
  APMSessionId m_id;
  APMEventLevel m_level;
  std::vector<std::string> m_filters;
public:
  CAPMSession(APMSessionId _id, APMEventLevel _level, const std::vector<std::string> &_filters) :
      m_id(_id), m_level(_level), m_filters(_filters)
  {
  }

  bool Match(const std::string &filter)
  {
    if (m_filters.empty()) {
      return true;
    }
    std::vector<std::string>::iterator iter;
    for (iter = m_filters.begin(); iter != m_filters.end(); ++iter) {
      if (*iter == filter) {
        return true;
      }
    }
    return false;
  }

  APMSessionId getId() const
  {
    return m_id;
  }

  APMEventLevel getLevel() const
  {
    return m_level;
  }

};

typedef std::vector<CAPMSession> CAPMSessionList;
typedef std::deque<std::string> CAPMEventList;

class CAPMWebEventDescription
{
private:
  std::string m_name;
  std::string m_parameters;
  bool m_refresh;
public:
  CAPMWebEventDescription(const char *name, const char *parameters) :
      m_name(name ? name : ""), m_parameters(parameters ? parameters : ""), m_refresh(true)
  {
  }

  const std::string GetName() const
  {
    return m_name;
  }

  const std::string GetParameters() const
  {
    return m_parameters;
  }

  bool doRefresh() const
  {
    return m_refresh;
  }

  void setRefresh(bool refresh)
  {
    m_refresh = refresh;
  }

};

typedef std::map<APMEventId, CAPMWebEventDescription> CAPMWebEventDescriptionMap;

class CAPMWebModule
{
private:
  std::string                 m_name;
  CAPMWebEventDescriptionMap  m_eventDescriptions;
  bool                        m_refresh;
  UInt64                      m_timestamp;

public:
  CAPMWebModule(const char *name) :
    m_name(name ? name :  "<unknown>"),
    m_refresh(true),
    m_timestamp(GetTimestamp())
  {
  }

  const std::string &getName() const
  {
    return m_name;
  }

  void AddEvent(APMEventId id, const char *name, const char *parameters)
  {
    m_eventDescriptions.insert(std::make_pair(id, CAPMWebEventDescription(name, parameters)));
    m_timestamp = GetTimestamp();
    m_refresh = true;
  }

  bool doRefresh() const
  {
    return m_refresh;
  }

  UInt64 getTimestamp()
  {
    return m_timestamp;
  }

  void setRefresh(bool refresh)
  {
    m_refresh = refresh;
    if (refresh) {
      CAPMWebEventDescriptionMap::iterator iter;
      for (iter = m_eventDescriptions.begin(); iter != m_eventDescriptions.end(); ++iter) {
        iter->second.setRefresh(true);
      }
    }
  }

  bool Write(char *&buffer, unsigned int &size, bool itemWritten, APMModuleHandle id)
  {
    bool success = WriteBuffer(buffer, size, '[') &&
           WriteBufferNumeric(buffer, size, id) &&
           WriteBuffer(buffer, size, ',') &&
           WriteBufferStringEscape(buffer, size, m_name) &&
           WriteBuffer(buffer, size, ",[");
    if (success) {
      bool first = true;
      size -= 2;
      bool written = true;
      CAPMWebEventDescriptionMap::iterator iter;
      for (iter = m_eventDescriptions.begin(); success && written && iter != m_eventDescriptions.end(); ++iter) {
        if (iter->second.doRefresh()) {
          char *p2 = buffer;
          unsigned int s2 = size;
          if (first) {
            first = false;
          }
          else {
            written = WriteBuffer(buffer, size, ',');
          }
          written = WriteBuffer(buffer, size, '[') &&
             WriteBufferNumeric(buffer, size, iter->first) &&
             WriteBuffer(buffer, size, ',') &&
             WriteBufferStringEscape(buffer, size, iter->second.GetName());
          if (written && !iter->second.GetParameters().empty()) {
             WriteBuffer(buffer, size, ',') &&
             WriteBufferStringEscape(buffer, size, iter->second.GetParameters());
          }
          written = written &&
             WriteBuffer(buffer, size, ']');
          if (written) {
            itemWritten = true;
            iter->second.setRefresh(false);
          }
          else {
            if (!itemWritten) {
              iter->second.setRefresh(false); // skip large entries
              written = true;
            }
            buffer = p2;
            size = s2;
          }
        }
      }
      if (written) {
        m_refresh = false;
      }
      size += 2;
      success = WriteBuffer(buffer, size, "]]");
    }
    return success;
  }

  void Refresh()
  {
    CAPMWebEventDescriptionMap::iterator iter;
    for (iter = m_eventDescriptions.begin(); iter != m_eventDescriptions.end(); ++iter) {
      iter->second.setRefresh(true);
    }
    m_refresh = true;
  }

};

typedef std::map<APMModuleHandle, CAPMWebModule> CAPMWebModuleMap;

class CAPMWebApplication
{
protected:
  struct EventBufferHeader {
    UInt32    size;
    UInt32    readptr;
    UInt32    writeptr;

    EventBufferHeader(UInt32 _size) :
      size(_size), readptr(0), writeptr(0)
    {
    }
  };

protected:
  APMApplicationHandle m_handle;
  std::string m_procid;
  std::string m_name;
  std::string m_id;
  std::string m_tiername;
  std::string m_tierid;
  bool m_refresh;
  UInt64 m_timestamp;
  UInt64 m_flushTimestamp;

  CAPMWebFilterValueMap m_filterValues;

  CAPMSessionList    m_sessions;
  std::string        m_guid;
  APMEventLevel      m_loglevel;
  bool               m_logdistributed;

  void              *m_eventBuffer;
  EventBufferHeader *m_eventBufferHeader;

  UInt32             m_eventReadPtr;

  bool m_eventdropped;
  std::set<APMModuleHandle> m_backupSentModules;
  std::set<APMModuleHandle> m_sentModules;
  UInt64 m_refreshRegistration;

#if defined APM_WEB_API_PERSIST
  IAPMWebPersistStore       *m_persistHandler;
  APMWebPersistStorageHandle m_persistStorageHandle;
#endif

public:
  CAPMWebApplication(APMApplicationHandle handle) :
    m_handle(handle),
    m_procid(), m_name(), m_id(), m_tiername(), m_tierid(),
    m_refresh(false),
    m_timestamp(GetTimestamp()), m_flushTimestamp(0),
    m_filterValues(), m_sessions(), m_guid(),
    m_loglevel(APM_EVENT_LEVEL_LOG), m_logdistributed(false),
    m_eventBuffer(NULL), m_eventBufferHeader(NULL),
    m_eventReadPtr(0),
    m_eventdropped(false),
    m_backupSentModules(), m_sentModules(),
    m_refreshRegistration(0)
#if defined APM_WEB_API_PERSIST
    , m_persistHandler(NULL), m_persistStorageHandle(NULL)
#endif
  {
    APMContext localcontext;
    _apmthread_t threadid = _getThreadId();
    UInt64 timestamp = GetTimestamp();
    memcpy(localcontext, (const void *)&timestamp, sizeof(timestamp));
    memcpy(localcontext + sizeof(timestamp), (const void *)&threadid, sizeof(threadid));
    std::stringstream param;
    WriteBase64(param, &localcontext, sizeof(localcontext));
    m_procid = param.str();
  }

  ~CAPMWebApplication()
  {
#if defined(APM_WEB_API_PERSIST)
    bool deleteFiles = false;
    if (m_eventBufferHeader && m_eventBufferHeader->readptr == m_eventBufferHeader->writeptr) {
      deleteFiles = true;
    }
    if (m_persistHandler && m_persistStorageHandle) {
      m_persistHandler->closefn(m_persistStorageHandle, deleteFiles);
      m_eventBuffer = NULL;
      m_eventBufferHeader = NULL;
    }
#endif
    if (m_eventBuffer) {
      free(m_eventBuffer);
    }
    if (m_eventBufferHeader) {
      delete(m_eventBufferHeader);
    }
  }

  void Init(const char *name, const char *id, const char *tiername, const char *tierid)
  {
    if (!m_eventBuffer) {
      m_eventBuffer = malloc(APM_WEB_API_DEFAULT_BUFFERSIZE);
      m_eventBufferHeader = new EventBufferHeader(APM_WEB_API_DEFAULT_BUFFERSIZE);
    }
    m_name.assign(name ? name : "");
    m_id.assign(id ? id : "");
    m_tiername.assign(tiername ? tiername : "");
    m_tierid.assign(tierid ? tierid : "");
    m_timestamp = GetTimestamp();
    m_refresh = true;
  }

  void InitPersistent(IAPMWebPersistStore *persistHandler)
  {
    m_persistHandler = persistHandler;
    if (persistHandler && persistHandler->createfn) {
      void *buffer = NULL;
      size_t size = 0;
      m_persistStorageHandle = persistHandler->createfn(&buffer, &size);
      if (m_persistStorageHandle) {
        if (buffer && size > sizeof(EventBufferHeader)) {
          m_eventBufferHeader = static_cast<EventBufferHeader*>(buffer);
          m_eventBufferHeader->size = (UInt32)(size - sizeof(EventBufferHeader));
          m_eventBufferHeader->readptr = 0;
          m_eventBufferHeader->writeptr = 0;
          m_eventBuffer = static_cast<char*>(buffer) + sizeof(EventBufferHeader);
        }
        else {
          persistHandler->closefn(m_persistStorageHandle, true);
          m_persistStorageHandle = NULL;
        }
      }
    }
  }

  void Clear()
  {
    m_name.clear();
    m_id.clear();
    m_tiername.clear();
    m_tierid.clear();
    m_filterValues.clear();
    m_refresh = false;
  }

  void PrepareSend()
  {
    if (m_eventReadPtr != m_eventBufferHeader->readptr || !m_backupSentModules.empty()) {
      m_backupSentModules.clear();
    }
    // refresh module and filter registration every 10 minutes
    UInt64 now = GetTimestamp();
    if (m_refreshRegistration == 0 || (now - m_refreshRegistration) > GetTimestampFrequency() * 600ULL) {
      m_sentModules.clear();
      m_refreshRegistration = now;
      CAPMWebFilterValueMap::iterator iter;
      for (iter = m_filterValues.begin(); iter != m_filterValues.end(); ++iter) {
        iter->second.setRefresh(true);
      }
    }
  }

  bool SendModule(APMModuleHandle module)
  {
    if (m_sentModules.find(module) == m_sentModules.end()) {
      m_backupSentModules.insert(module);
      return true;
    }
    else {
      return false;
    }
  }

  void RegisterFilterValue(const char *value, const char *description)
  {
    m_timestamp = GetTimestamp();
    m_refresh = true;
    if (value) {
      CAPMWebFilterValueMap::iterator iter = m_filterValues.find(value);
      if (iter == m_filterValues.end()) {
        m_filterValues.insert(std::make_pair(value, CAPMFilterValue(description)));
      }
      else {
        iter->second.setDescription(description);
      }
    }
    else {
      CAPMWebFilterValueMap::iterator iter = m_filterValues.find(std::string());
      if (iter == m_filterValues.end()) {
        m_filterValues.insert(std::make_pair(std::string(), description ? description : ""));
      }
      else {
        iter->second = description ? description : "";
      }
    }
  }

  void MatchFilter(const char *filter, APMEventLevel &eventLevel, std::vector<APMSessionId> &sessions)
  {
    std::string sfilter(toLowerString(filter ? filter : ""));
    CAPMSessionList::iterator iter;
    for (iter = m_sessions.begin(); iter != m_sessions.end(); ++iter) {
      if (iter->Match(sfilter)) {
        sessions.push_back(iter->getId());
        if (eventLevel < iter->getLevel()) {
          eventLevel = iter->getLevel();
        }
      }
    }
  }

  APMEventLevel GetLogLevel() const
  {
    return m_loglevel;
  }

  bool GetLogDistributed() const
  {
    return m_logdistributed;
  }

  bool WriteRegistration(char *&buffer, unsigned int &size)
  {
    bool success =
      WriteBufferStringEscape(buffer, size, m_procid) &&
      WriteBuffer(buffer, size, ',') &&
      WriteBufferStringEscape(buffer, size, m_name) &&
      WriteBuffer(buffer, size, ',') &&
      WriteBufferStringEscape(buffer, size, m_id) &&
      WriteBuffer(buffer, size, ',') &&
      WriteBufferStringEscape(buffer, size, m_tiername) &&
      WriteBuffer(buffer, size, ',') &&
      WriteBufferStringEscape(buffer, size, m_tierid);
    return success;
  }

  bool HasData() const
  {
    return m_refreshRegistration == 0 || m_eventBufferHeader->readptr != m_eventBufferHeader->writeptr;
  }

  void SetData(
    const void *buffer,
    unsigned int size)
  {
    m_eventBufferHeader->readptr = m_eventReadPtr;
    if (m_eventBufferHeader->readptr == m_eventBufferHeader->writeptr) {
      m_eventBufferHeader->readptr = 0;
      m_eventBufferHeader->writeptr = 0;
      m_eventReadPtr = 0;
    }
    m_sentModules.insert(m_backupSentModules.begin(), m_backupSentModules.end());
    m_backupSentModules.clear();

    _JSONParser parser((const char *)buffer, size);
    bool success = true;

#define EXPECT_TOKEN(token)                     \
  if (success && parser.getToken() != token) {  \
    success = false;                            \
  }

#define EXPECT_NEXT_TOKEN(token)                \
  if (success && parser.getNext() != token) {   \
    success = false;                            \
  }

    EXPECT_NEXT_TOKEN(json_object_start);
    while (success && parser.getNext() == json_string_token) {
      if (parser.getValue() == "sessions") {
        EXPECT_NEXT_TOKEN(json_colon_separator);
        EXPECT_NEXT_TOKEN(json_array_start);
        m_sessions.clear();
        if (parser.getNext() != json_array_end) {
          while (success) {
            success = parser.getToken() == json_object_start;
            APMSessionId sessionid = 0;
            APMEventLevel level = 0;
            std::vector<std::string> filters;
            while (success) {
              if (parser.getNext() != json_string_token) {
                break;
              }
              if (parser.getValue() == "id") {
                EXPECT_NEXT_TOKEN(json_colon_separator);
                EXPECT_NEXT_TOKEN(json_number_token);
                if (success) {
                  sessionid = (APMSessionId)parser.getNumber();
                }
              }
              else if (parser.getValue() == "level") {
                EXPECT_NEXT_TOKEN(json_colon_separator);
                EXPECT_NEXT_TOKEN(json_number_token);
                if (success) {
                  level = (APMEventLevel)parser.getNumber();
                }
              }
              else if (parser.getValue() == "filters") {
                EXPECT_NEXT_TOKEN(json_colon_separator);
                EXPECT_NEXT_TOKEN(json_array_start);
                while (success && parser.getNext() == json_string_token) {
                  filters.push_back(toLowerString(parser.getValue()));
                  if (parser.getNext() == json_array_end) {
                    break;
                  }
                }
              }
              else {
                EXPECT_NEXT_TOKEN(json_colon_separator);
                success = false;
                break;
              }
              if (parser.getNext() == json_object_end) {
                break;
              }
            }
            if (sessionid != 0) {
              m_sessions.push_back(CAPMSession(sessionid, level, filters));
            }
            if (parser.getNext() == json_array_end) {
              break;
            }
            success = parser.getToken() == json_comma_separator;
            parser.getNext();
          }
        }
      }
      else if (parser.getValue() == "loglevel") {
        EXPECT_NEXT_TOKEN(json_colon_separator);
        EXPECT_NEXT_TOKEN(json_number_token);
        if (success) {
          m_loglevel = (APMEventLevel)parser.getNumber();
        }
      }
      else if (parser.getValue() == "logdistributed") {
        EXPECT_NEXT_TOKEN(json_colon_separator);
        if (success) {
          switch (parser.getNext()) {
          case json_bool_true:
            m_logdistributed = true;
            break;
          case json_bool_false:
            m_logdistributed = false;
            break;
          default:
            break;
          }
        }
      }
      else if (parser.getValue() == "GUID") {
        EXPECT_NEXT_TOKEN(json_colon_separator);
        EXPECT_NEXT_TOKEN(json_string_token);
        if (success) {
          m_guid = parser.getValue();
        }
      }
      else {
        // skip unknown nodes
        EXPECT_NEXT_TOKEN(json_colon_separator);
        if (success) {
          parser.getNext();
        }
      }
      if (parser.getNext() == json_object_end) {
        break;
      }
      if (parser.getToken() != json_comma_separator) {
        success = false;
        break;
      }
    }
    EXPECT_NEXT_TOKEN(json_end);
#undef EXPECT_TOKEN
#undef EXPECT_NEXT_TOKEN
  }

  bool WriteFilter(char *&p, unsigned int &s, bool &itemWritten)
  {
    bool success = WriteBuffer(p, s, ',');
    bool firstFilter = true;
    CAPMWebFilterValueMap::iterator iter;
    for (iter = m_filterValues.begin(); success && iter != m_filterValues.end(); ++iter) {
      if (iter->second.doRefresh()) {
        bool fsuccess = true;
        char *p1 = p;
        unsigned int s1 = s;
        bool wasFirstFilter = firstFilter;
        if (firstFilter) {
          fsuccess = WriteBuffer(p, s, '{') && s > 1;
          if (fsuccess) {
            s--;
          }
        }
        else {
          fsuccess = WriteBuffer(p, s, ',');
        }
        fsuccess = fsuccess &&
          WriteBufferStringEscape(p, s, iter->first) &&
          WriteBuffer(p, s, ':') &&
          WriteBufferStringEscape(p, s, iter->second.getDescription());
        if (fsuccess) {
          itemWritten = true;
          iter->second.setRefresh(false);
        }
        else {
          p = p1;
          s = s1;
          firstFilter = wasFirstFilter;
          if (!itemWritten) {
            iter->second.setRefresh(false); // skip large item
          }
        }
      }
    }
    if (success) {
      if (!firstFilter) {
        s++;
        WriteBuffer(p, s, '}');
      }
      else {
        WriteBuffer(p, s, "null");
      }
    }
    return success;
  }

  bool WriteModules(char *&p, unsigned int &s, bool &itemWritten,
    CAPMWebModuleMap &modules,
    bool writeFile)
  {
    bool firstModule = true;
    CAPMWebModuleMap::iterator moduleIter;
    for (moduleIter = modules.begin(); moduleIter != modules.end(); ++moduleIter) {
      if (writeFile || SendModule(moduleIter->first)) {
        moduleIter->second.setRefresh(true);
      }
      if (moduleIter->second.doRefresh()) {
        bool success = true;
        char *p1 = p;
        unsigned int s1 = s;
        bool wasFirstModule = firstModule;
        if (firstModule) {
          firstModule = false;
          success = WriteBuffer(p, s, ",[") && s > 1;
          if (success) {
            s--;
          }
        }
        else {
          success = WriteBuffer(p, s, ',');
        }
        success = success &&
          moduleIter->second.Write(p, s, itemWritten, moduleIter->first);
        if (!success) {
          p = p1;
          s = s1;
          firstModule = wasFirstModule;
        }
      }
    }
    if (!firstModule) {
      s++;
      WriteBuffer(p, s, ']');
    }
    return true;
  }

  void WriteToEventBuffer(const void *buffer, UInt32 size)
  {
    const char *p = (char*)buffer;
    UInt32 blockSize = m_eventBufferHeader->size - m_eventBufferHeader->writeptr;
    if (size >= blockSize) {
      if (blockSize) {
        memcpy(static_cast<char*>(m_eventBuffer) + m_eventBufferHeader->writeptr, p, blockSize);
        size -= blockSize;
        p += blockSize;
      }
      m_eventBufferHeader->writeptr = 0;
    }
    if (size) {
      memcpy(static_cast<char*>(m_eventBuffer) + m_eventBufferHeader->writeptr, p, size);
      m_eventBufferHeader->writeptr += size;
    }
  }

  void ReadFromEventBuffer(UInt32 &readptr, void *buffer, UInt32 size)
  {
    char *p = (char*)buffer;
    UInt32 blockSize = m_eventBufferHeader->size - readptr;
    if (size >= blockSize) {
      if (blockSize) {
        memcpy(p, static_cast<char*>(m_eventBuffer) + readptr, blockSize);
        size -= blockSize;
        p += blockSize;
      }
      readptr = 0;
    }
    if (size) {
      memcpy(p, static_cast<char*>(m_eventBuffer) + readptr, size);
      readptr += size;
    }
  }

  void ReadFromEventBuffer(UInt32 &readptr, std::string &value, UInt32 size)
  {
    value.clear();
    value.reserve(size + 1);
    UInt32 blockSize = m_eventBufferHeader->size - readptr;
    if (size >= blockSize) {
      if (blockSize) {
        value.append(static_cast<char*>(m_eventBuffer) + readptr, blockSize);
        size -= blockSize;
      }
      readptr = 0;
    }
    if (size) {
      value.append(static_cast<char*>(m_eventBuffer) + readptr, size);
      readptr += size;
    }
  }

  void QueueEvent(const std::string &value)
  {
    if (m_eventBuffer) {
      UInt32 size = (UInt32)value.length();
      UInt32 freeSize = m_eventBufferHeader->size - m_eventBufferHeader->writeptr + m_eventBufferHeader->readptr - 1;
      if (freeSize >= m_eventBufferHeader->size) {
        freeSize -= m_eventBufferHeader->size;
      }
      if (freeSize > size + sizeof(size)) {
        WriteToEventBuffer(&size, sizeof(size));
        WriteToEventBuffer(value.data(), size);
      }
      else {
        m_eventdropped = true;
      }
    }
  }

  bool ReadEvent(UInt32 &readptr, std::string &value)
  {
    bool success = false;
    UInt32 rp = readptr;
    UInt32 availSize;
    if (m_eventBufferHeader->writeptr >= rp) {
      availSize = m_eventBufferHeader->writeptr - rp;
    }
    else {
      availSize = m_eventBufferHeader->size + m_eventBufferHeader->writeptr - rp;
    }
    if (availSize >= sizeof(UInt32)) {
      UInt32 size = 0;
      ReadFromEventBuffer(rp, &size, sizeof(size));
      availSize -= sizeof(size);
      if (availSize >= size) {
        ReadFromEventBuffer(rp, value, size);
        readptr = rp;
        success = true;
      }
      else {
        readptr = m_eventBufferHeader->writeptr;
      }
    }
    else {
      readptr = m_eventBufferHeader->writeptr;
    }
    return success;
  }

  void WriteEvents(char *&p, unsigned int &s, bool &itemWritten)
  {
    bool firstEvent = true;
    std::string value;
    m_eventReadPtr = m_eventBufferHeader->readptr;
    while (ReadEvent(m_eventReadPtr, value)) {
      bool success = true;
      char *p1 = p;
      unsigned int s1 = s;
      bool wasFirstEvent = firstEvent;
      if (firstEvent) {
        firstEvent = false;
        success = WriteBuffer(p, s, ",[") && s > 1;
        if (success) {
          s--;
        }
      }
      else {
        success = WriteBuffer(p, s, ',');
      }
      success = success && WriteBuffer(p, s, '[') &&
          WriteBuffer(p, s, value) &&
          WriteBuffer(p, s, ']');
      if (success) {
        itemWritten = true;
      }
      else {
        p = p1;
        s = s1;
        firstEvent = wasFirstEvent;
        if (itemWritten) {
          break;
        }
      }
    }
    if (!firstEvent) {
      s++;
      WriteBuffer(p, s, ']');
    }
  }

  int GetData(
    void *buffer,
    unsigned int *size,
    CAPMWebModuleMap &modules,
    bool writeFile = false)
  {
    bool result = 0;
    PrepareSend();
    char *p = (char *)buffer;
    unsigned int s = *size;
    bool itemWritten = false;
    bool success = true;
    if (!writeFile) {
      success = WriteBuffer(p, s, "{\"apm\":[") && s > 2;
      if (success) {
        s -= 2;
        char stringbuffer[30];
  #if defined(APM_WIN)
        SYSTEMTIME time;
        GetSystemTime(&time);
        sprintf_s(stringbuffer, "%04u-%02u-%02uT%02u:%02u:%02u.%03u",
          time.wYear, time.wMonth, time.wDay, time.wHour,
          time.wMinute, time.wSecond, time.wMilliseconds);
  #elif defined(APM_UNIX)
        tzset();
        struct timeval tv;
        gettimeofday(&tv, NULL);
        tv.tv_sec += (timezone - (daylight != 0 ? 3600 : 0));
        struct tm loctime;
        localtime_r ((time_t *)&tv.tv_sec, &loctime);
        sprintf(stringbuffer, "%04u-%02u-%02uT%02u:%02u:%02u.%03u\n",
          loctime.tm_year + 1900, loctime.tm_mon + 1, loctime.tm_mday,
          loctime.tm_hour, loctime.tm_min, loctime.tm_sec,
          (unsigned)(tv.tv_usec / 1000));
  #endif
        success =
          WriteBuffer(p, s, '[') &&
          WriteBufferStringEscape(p, s, PRODUCT_VERSION_STR) &&
          WriteBuffer(p, s, ',') && WriteBufferNumeric(p, s, GetTimestamp()) &&
          WriteBuffer(p, s, ',') && WriteBufferNumeric(p, s, GetTimestampFrequency()) &&
          WriteBuffer(p, s, ',') && WriteBufferStringEscape(p, s, stringbuffer);
          WriteBuffer(p, s, ',') && WriteBufferStringEscape(p, s, m_guid);
          WriteBuffer(p, s, ']');
      }
    }
    else {
      success = WriteBuffer(p, s, "[") && s > 1;
      if (success) {
        s -= 1;
        success =
          WriteBuffer(p, s, '[') && WriteBufferStringEscape(p, s, PRODUCT_VERSION_STR) &&
          WriteBuffer(p, s, ',') && WriteBufferStringEscape(p, s, m_guid) &&
          WriteBuffer(p, s, ']');
      }
    }
    success = success &&
      WriteBuffer(p, s, ',') &&
      WriteBuffer(p, s, '[') &&
      WriteRegistration(p, s) &&
      WriteFilter(p, s, itemWritten) &&
      WriteModules(p,s, itemWritten, modules, writeFile) &&
      WriteBuffer(p, s, ']');

    if (success) {
      if (!writeFile) {
        WriteEvents(p, s, itemWritten);
        s += 2;
        success = success && WriteBuffer(p, s, "]}");
      }
      else {
        s += 1;
        success = success && WriteBuffer(p, s, "]");
      }
    }
    if (success) {
      *size -= s;
      result = 1;
    }
    return result;
  }

  void Flush(CAPMWebModuleMap &modules)
  {
    if (m_persistHandler != NULL && m_persistStorageHandle != NULL) {
      UInt64 timestamp = m_timestamp;
      CAPMWebModuleMap::iterator moduleIter;
      for (moduleIter = modules.begin(); moduleIter != modules.end(); ++moduleIter) {
        timestamp = 16777619U * timestamp ^ moduleIter->second.getTimestamp();
      }
      if (m_flushTimestamp != timestamp) {
        unsigned int size = 0x10000;
        void *buffer = malloc(size);
        while (buffer) {
          unsigned int sendSize = size;
          if (GetData(buffer, &sendSize, modules, true)) {
            if (m_persistHandler->writeappdatafn) {
              (*m_persistHandler->writeappdatafn)(m_persistStorageHandle, buffer, sendSize);
            }
            break;
          }
          else {
            size *= 2;
            free(buffer);
            buffer = malloc(size);
          }
        }
        if (buffer) {
          free(buffer);
        }
        m_flushTimestamp = timestamp;
      }
    }
  }


};

class CAPMWebRecoveryApplication :
  public CAPMWebApplication
{
private:
  CAPMWebModuleMap m_modules;

public:
  CAPMWebRecoveryApplication() :
    CAPMWebApplication(0)
  {
  }

  ~CAPMWebRecoveryApplication()
  {
    m_eventBufferHeader = NULL;
    m_eventBuffer = NULL;
  }

  bool InitRecovery(void *buffer, size_t size, void *data, size_t datasize)
  {
    bool success = buffer != NULL && size && data != NULL && datasize > sizeof(EventBufferHeader);
    if (success) {
      m_eventBufferHeader = static_cast<EventBufferHeader*>(data);
      m_eventBuffer = static_cast<char*>(data) + sizeof(EventBufferHeader);
      success = m_eventBufferHeader->size <= datasize - sizeof(EventBufferHeader) &&
                m_eventBufferHeader->readptr <= datasize - sizeof(EventBufferHeader) &&
                m_eventBufferHeader->writeptr <= datasize - sizeof(EventBufferHeader);
    }
    if (success && buffer) {
      _JSONParser parser((const char *)buffer, (unsigned int)size);
      success =
        parser.getNext() == json_array_start &&
        parser.getNext() == json_array_start &&
        parser.getNext() == json_string_token && // ignore version
        parser.getNext() == json_comma_separator &&
        parser.getNext() == json_string_token;
      if (success) {
        m_guid = parser.getValue();
      }
      success = success &&
        parser.getNext() == json_array_end &&
        parser.getNext() == json_comma_separator &&
        parser.getNext() == json_array_start &&
        parser.getNext() == json_string_token;
      if (success) {
        m_procid = parser.getValue();
        success =
          parser.getNext() == json_comma_separator &&
          parser.getNext() == json_string_token;
      }
      if (success) {
        m_name = parser.getValue();
        success =
          parser.getNext() == json_comma_separator &&
          parser.getNext() == json_string_token;
      }
      if (success) {
        m_id = parser.getValue();
        success =
          parser.getNext() == json_comma_separator &&
          parser.getNext() == json_string_token;
      }
      if (success) {
        m_tiername = parser.getValue();
        success =
          parser.getNext() == json_comma_separator &&
          parser.getNext() == json_string_token;
      }
      if (success) {
        m_tierid = parser.getValue();
      }
      success = success &&
        parser.getNext() == json_comma_separator;
      if (success) {
        std::string key;
        switch (parser.getNext()) {
        case json_object_start:
          do {
            switch (parser.getNext()) {
            case json_object_end:
              break;
            case json_string_token:
              key = parser.getValue();
              success =
                parser.getNext() == json_colon_separator &&
                parser.getNext() == json_string_token;
              if (success) {
                m_filterValues.insert(std::make_pair(key, CAPMFilterValue(parser.getValue().c_str())));
              }
              break;
            default:
              success = false;
            }
          } while (success && parser.getNext() == json_comma_separator);
          success = success &&
            parser.getToken() == json_object_end;
          break;
        case json_null:
          break;
        default:
          success = false;
          break;
        }
      }
      if (success && parser.getNext() == json_comma_separator) {
        success = parser.getNext() == json_array_start;
        if (success) {
          CAPMWebModuleMap::iterator moduleIter = m_modules.end();
          do {
            success =
              parser.getNext() == json_array_start &&
              parser.getNext() == json_number_token;
            if (success) {
              APMModuleHandle moduleid = (APMModuleHandle)parser.getNumber();
              success =
                parser.getNext() == json_comma_separator &&
                parser.getNext() == json_string_token;
              if (success) {
                std::string moduleName = parser.getValue();
                moduleIter = m_modules.insert(std::make_pair(moduleid, CAPMWebModule(moduleName.c_str()))).first;
                success =
                  parser.getNext() == json_comma_separator &&
                  parser.getNext() == json_array_start;
              }
              if (success) {
                do {
                  success =
                    parser.getNext() == json_array_start &&
                    parser.getNext() == json_number_token;
                  if (success) {
                    UInt64 eventId = (UInt64)parser.getNumber();
                    success =
                      parser.getNext() == json_comma_separator &&
                      parser.getNext() == json_string_token;
                    if (success) {
                      std::string eventName = parser.getValue();
                      std::string eventParam;
                      if (parser.getNext() == json_comma_separator) {
                        success = parser.getNext() == json_string_token;
                        if (success) {
                          eventParam = parser.getValue();
                          parser.getNext();
                        }
                      }
                      if (success) {
                        moduleIter->second.AddEvent(eventId, eventName.c_str(), eventParam.c_str());
                        success =
                          parser.getToken() == json_array_end;
                      }
                    }
                  }
                } while (success && parser.getNext() == json_comma_separator);
                success = success &&
                  parser.getToken() == json_array_end;
              }
            }
          } while (success && parser.getNext() == json_comma_separator);
          success = success &&
            parser.getToken() == json_array_end;
        }
      }
      //FIXME: Cppcheck duplicateExpression: Same expression on both sides of '&&'.
      success = success &&
        parser.getToken() == json_array_end &&
        parser.getToken() == json_array_end;
    }
    return success;
  }

  int GetData(
    void *buffer,
    unsigned int *size)
  {
    return CAPMWebApplication::GetData(buffer, size, m_modules, false);
  }

  void ConfirmData()
  {
    m_eventBufferHeader->readptr = m_eventReadPtr;
    m_sentModules.insert(m_backupSentModules.begin(), m_backupSentModules.end());
    m_backupSentModules.clear();
  }

};

static APM_ATOMICTYPE volatile g_context_cs = 0;

class CAPMWebThreadStatus
{
public:
  APMApplicationHandle m_application;
  APMEventLevel        m_eventlevel;

  CAPMWebThreadStatus(APMApplicationHandle application, APMEventLevel eventlevel) :
    m_application(application), m_eventlevel(eventlevel)
  {
  }
};

typedef std::map<APMApplicationHandle, CAPMWebApplication> CAPMWebApplicationMap;

class CAPMWebContext
{
private:
  CAPMWebApplicationMap m_applications;
  CAPMWebModuleMap      m_modules;

  APMModuleHandle       m_modulehandle;
  APMApplicationHandle  m_applicationhandle;
  APMTransactionHandle  m_txhandle;
  APMLogicalClock       m_logicalclock;

  typedef std::map<_apmthread_t, APMTransactionHandle> _ThreadTransactionMap;
  _ThreadTransactionMap m_threadTransactions;

  typedef std::map<APMTransactionHandle, CAPMWebThreadStatus> TransactionStatusMap;
  TransactionStatusMap  m_txEventLevels;

  std::string           m_directory;
  UInt32                m_bufferMaxSize;
  UInt64                m_checkPersistentTime;

#if defined APM_WEB_API_PERSIST
  IAPMWebPersistStore       *m_persistHandler;
#endif

private:

  static void Enter()
  {
    while (_apm_interlocked_exchange(&g_context_cs, -1) != 0) {
      _apm_yield();
    }
  }

  static void Leave()
  {
    g_context_cs = 0;
  }

private:
  void WriteItemHeader(std::stringstream &stream, UInt16 entryid, APMTransactionHandle tx)
  {
    stream <<
      entryid <<
      ',' << GetTimestamp() <<
      ',' << m_logicalclock++ <<
      ',' << tx;
  }

public:
  CAPMWebContext() :
      m_applications(), m_modules(),
      m_modulehandle(0), m_applicationhandle(0), m_txhandle(1), m_logicalclock(0),
      m_threadTransactions(), m_txEventLevels(),
      m_bufferMaxSize(0), m_checkPersistentTime(0)
#if defined(APM_WEB_API_PERSIST)
      , m_persistHandler(NULL)
#endif
  {
#if defined(APM_WIN)
    GetTimestampFrequency();
#endif
  }

  APMTransactionHandle GetThread()
  {
    _apmthread_t threadid = _getThreadId();
    APMTransactionHandle tx = 0;
    Enter();
    _ThreadTransactionMap::iterator iter = m_threadTransactions.find(threadid);
    if (iter != m_threadTransactions.end()) {
      tx = iter->second;
    }
    Leave();
    return tx;
  }

  void SetThread(APMTransactionHandle tx)
  {
    _apmthread_t threadid = _getThreadId();
    Enter();
    _ThreadTransactionMap::iterator iter = m_threadTransactions.find(threadid);
    if (iter != m_threadTransactions.end()) {
      iter->second = tx;
    }
    else {
      m_threadTransactions.insert(std::make_pair(threadid, tx));
    }
    Leave();
  }

  void ClearThread()
  {
    _apmthread_t threadid = _getThreadId();
    Enter();
    m_threadTransactions.erase(threadid);
    Leave();
  }

public:

  void SetPersistHandler(IAPMWebPersistStore *handler)
  {
    m_persistHandler = handler;
  }

  void Flush()
  {
    CAPMWebApplicationMap::iterator applicationIter;
    for (applicationIter = m_applications.begin(); applicationIter != m_applications.end(); ++applicationIter) {
      applicationIter->second.Flush(m_modules);
    }
  }

private:

  bool doPersist()
  {
    return m_persistHandler != NULL;
  }

public:

  APMApplicationHandle RegisterApplication(const char *name, const char *id, const char *tiername, const char *tierid)
  {
    Enter();
    APMApplicationHandle appHandle = m_applicationhandle++;
    CAPMWebApplicationMap::iterator appIter = m_applications.insert(std::make_pair(appHandle, CAPMWebApplication(appHandle))).first;
#if defined(APM_WEB_API_PERSIST)
    if (doPersist()) {
      appIter->second.InitPersistent(m_persistHandler);
    }
#endif
    appIter->second.Init(name, id, tiername, tierid);
    Leave();
    return appHandle;
  }

  void UnregisterApplication(APMApplicationHandle app)
  {
    Enter();
    m_applications.erase(app);
    Leave();
  }

  APMModuleHandle RegisterModule(const char *name)
  {
    Enter();
    CAPMWebModuleMap::iterator iter;
    for (iter = m_modules.begin(); iter != m_modules.end(); ++iter) {
      if ((name && iter->second.getName() == name) || (!name && iter->second.getName().empty())) {
        APMModuleHandle handle = iter->first;
        Leave();
        return handle;
      }
    }
    APMModuleHandle handle = m_modulehandle++;
    m_modules.insert(std::make_pair(handle, CAPMWebModule(name)));
    Leave();
    return handle;
  }

  void UnregisterModule(APMModuleHandle handle)
  {
    Enter();
    m_modules.erase(handle);
    Leave();
  }

  void RegisterFilterValue(APMApplicationHandle app, const char *value, const char *description)
  {
    UNREFERENCED_PARAMETER(app);
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      appIter->second.RegisterFilterValue(value, description);
    }
    Leave();
  }

  void ModuleRegisterEvent(APMModuleHandle module, APMEventId eventid, const char *name, const char *parameters)
  {
    Enter();
    CAPMWebModuleMap::iterator iter = m_modules.find(module);
    if (iter != m_modules.end()) {
      iter->second.AddEvent(eventid, name, parameters);
    }
    Leave();
  }

  APMTransactionHandle CreateContext(APMTransactionHandle thread, APMApplicationHandle app, const char *filtervalue)
  {
    UNREFERENCED_PARAMETER(app);
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
      APMEventLevel eventLevel = appIter->second.GetLogLevel();
      if (iter != m_txEventLevels.end()) {
        eventLevel = iter->second.m_eventlevel;
      }
      else {
        thread = m_txhandle++;
        iter = m_txEventLevels.insert(std::make_pair(thread, CAPMWebThreadStatus(app, eventLevel))).first;
      }
      std::vector<APMSessionId> sessions;
      appIter->second.MatchFilter(filtervalue, eventLevel, sessions);
      iter->second.m_eventlevel = eventLevel;
      std::stringstream stream;
      WriteItemHeader(stream, SWT_ENTRY_CREATE_CONTEXT, thread);
      WriteItemParamString(stream, filtervalue);
      if (!sessions.empty()) {
        WriteItemParamArrayNumeric(stream, sessions);
      }
      appIter->second.QueueEvent(stream.str());
    }
    Leave();
    return thread;
  }

  int GetContext(APMTransactionHandle thread, APMContext *context)
  {
    if (context == NULL) return 0;
    int result = 0;
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end()) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        APMEventLevel eventLevel = iter->second.m_eventlevel;
        if (eventLevel > APM_EVENT_LEVEL_LOG || appIter->second.GetLogDistributed()) {
          APMContext localcontext;
          _apmthread_t threadid = _getThreadId();
          UInt64 timestamp = GetTimestamp();
          UInt64 xtid = threadid ^ timestamp;
          memcpy(localcontext, (const void *)&xtid, sizeof(xtid));
          memcpy(localcontext + sizeof(xtid), (const void *)&m_logicalclock, sizeof(APMLogicalClock));
          localcontext[15] = eventLevel;
          memcpy(context, localcontext, sizeof(APMContext));
          std::stringstream stream;
          WriteItemHeader(stream, SWT_ENTRY_GET_CONTEXT, thread);
          WriteItemParamBase64(stream, &localcontext, sizeof(APMContext));
          appIter->second.QueueEvent(stream.str());
          result = 1;
        }
      }
    }
    Leave();
    return result;
  }

  APMTransactionHandle AttachContext(APMTransactionHandle thread, APMApplicationHandle app, const APMContext *context)
  {
    UNREFERENCED_PARAMETER(app);
    if (context == NULL) return thread;
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
      APMEventLevel eventLevel = appIter->second.GetLogLevel();
      if (iter != m_txEventLevels.end()) {
        eventLevel = iter->second.m_eventlevel;
      }
      else {
        thread = m_txhandle++;
        iter = m_txEventLevels.insert(std::make_pair(thread, CAPMWebThreadStatus(app, eventLevel))).first;
      }
      if ((*context)[15] > eventLevel) {
        iter->second.m_eventlevel = (*context)[15];
      }
      std::stringstream stream;
      WriteItemHeader(stream, SWT_ENTRY_ATTACH_CONTEXT, thread);
      WriteItemParamBase64(stream, context, sizeof(APMContext));
      appIter->second.QueueEvent(stream.str());
    }
    Leave();
    return thread;
  }

  void SyncContext(APMTransactionHandle thread, const APMContext *context)
  {
    (void)thread;
    (void)context;
  }

  void ReleaseContext(APMTransactionHandle thread)
  {
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end()) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        std::stringstream stream;
        WriteItemHeader(stream, SWT_ENTRY_RELEASE_CONTEXT, thread);
        appIter->second.QueueEvent(stream.str());
        m_txEventLevels.erase(iter);
      }
    }
    Leave();
  }

  int GetSyncMark(APMTransactionHandle thread, APMSyncMark *syncmark)
  {
    if (syncmark == NULL) return 0;
    Enter();
    int result = 0;
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end()) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        APMEventLevel eventLevel = iter->second.m_eventlevel;
        if (eventLevel > APM_EVENT_LEVEL_LOG || appIter->second.GetLogDistributed()) {
          APMSyncMark localsyncmark;
          _apmthread_t threadid = _getThreadId();
          UInt64 timestamp = GetTimestamp();
          UInt64 xtid = threadid ^ timestamp;
          memcpy(localsyncmark, (const void *)&xtid, sizeof(xtid));
          memcpy(localsyncmark + sizeof(xtid), (const void *)&m_logicalclock, sizeof(APMLogicalClock));
          memcpy(syncmark, localsyncmark, sizeof(APMSyncMark));
          std::stringstream stream;
          WriteItemHeader(stream, SWT_ENTRY_GET_SYNC_MARK, thread);
          WriteItemParamBase64(stream, &localsyncmark, sizeof(APMSyncMark));
          appIter->second.QueueEvent(stream.str());
          result = 1;
        }
      }
    }
    Leave();
    return result;
  }

  void SetSyncMark(APMTransactionHandle thread, const APMSyncMark *syncmark)
  {
    if (syncmark == NULL) return;
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end()) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        std::stringstream stream;
        WriteItemHeader(stream, SWT_ENTRY_SET_SYNC_MARK, thread);
        WriteItemParamBase64(stream, syncmark, sizeof(APMSyncMark));
        appIter->second.QueueEvent(stream.str());
      }
    }
    Leave();
  }

  int HasActiveContext(APMTransactionHandle thread, APMEventLevel level)
  {
    int result = 0;
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end()) {
      result = iter->second.m_eventlevel >= level ? 1 : 0;
    }
    Leave();
    return result;
  }

  APMTransactionHandle Report(APMTransactionHandle thread, APMApplicationHandle app,
    const char *filtervalue,
    const char *reportkey,
    const char *description)
  {
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      std::stringstream stream;
      WriteItemHeader(stream, SWT_ENTRY_REPORT, thread);
      WriteItemParamString(stream, filtervalue);
      WriteItemParamString(stream, reportkey);
      if (description) {
        WriteItemParamString(stream, description);
      }
      appIter->second.QueueEvent(stream.str());
    }
    return thread;
  }

  APMTransactionHandle ReportValue(APMTransactionHandle thread, APMApplicationHandle app,
    const char *reportkey,
    const char *key,
    const char *value)
  {
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      std::stringstream stream;
      WriteItemHeader(stream, SWT_ENTRY_REPORT_VALUE, thread);
      WriteItemParamString(stream, reportkey);
      WriteItemParamString(stream, key);
      WriteItemParamString(stream, value);
      appIter->second.QueueEvent(stream.str());
    }
    Leave();
    return thread;
  }

  void Event(
    APMTransactionHandle thread,
    APMModuleHandle module,
    APMEventId eventid,
    APMEventLevel level,
    APMEventFlags flags)
  {
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end() && iter->second.m_eventlevel >= level) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        std::stringstream stream;
        WriteItemHeader(stream, SWT_ENTRY_EVENT, thread);
        WriteItemParamNumeric(stream, module);
        WriteItemParamNumeric(stream, eventid);
        WriteItemParamNumeric(stream, (unsigned)level);
        WriteItemParamNumeric(stream, (unsigned)flags);
        appIter->second.QueueEvent(stream.str());
      }
    }
    Leave();
  }

  void EventStr(
    APMTransactionHandle thread,
    APMModuleHandle module,
    APMEventId eventid,
    APMEventLevel level,
    APMEventFlags flags,
    const char *param)
  {
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end() && iter->second.m_eventlevel >= level) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        std::stringstream stream;
        WriteItemHeader(stream, SWT_ENTRY_EVENT, thread);
        WriteItemParamNumeric(stream, module);
        WriteItemParamNumeric(stream, eventid);
        WriteItemParamNumeric(stream, (unsigned)level);
        WriteItemParamNumeric(stream, (unsigned)flags);
        WriteItemParamString(stream, param);
        appIter->second.QueueEvent(stream.str());
      }
    }
    Leave();
  }

  void EventParam(
    APMTransactionHandle thread,
    APMModuleHandle module,
    APMEventId eventid,
    APMEventLevel level,
    APMEventFlags flags,
    std::stringstream *param)
  {
    Enter();
    TransactionStatusMap::iterator iter = m_txEventLevels.find(thread);
    if (iter != m_txEventLevels.end() && iter->second.m_eventlevel >= level) {
      CAPMWebApplicationMap::iterator appIter = m_applications.find(iter->second.m_application);
      if (appIter != m_applications.end()) {
        std::stringstream stream;
        WriteItemHeader(stream, SWT_ENTRY_EVENT, thread);
        WriteItemParamNumeric(stream, module);
        WriteItemParamNumeric(stream, eventid);
        WriteItemParamNumeric(stream, (unsigned)level);
        WriteItemParamNumeric(stream, (unsigned)flags);
        if (param) {
          stream << param->str();
        }
        appIter->second.QueueEvent(stream.str());
      }
    }
    Leave();
  }

  int GetData(
    APMApplicationHandle app,
    void *buffer,
    unsigned int *size)
  {
    int result = 0;
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      CAPMWebApplication &apmapp = appIter->second;
      result = apmapp.GetData(buffer, size, m_modules);
      apmapp.Flush(m_modules);
    }
    Leave();
    return result;
  }

  int HasData(APMApplicationHandle app)
  {
    int result = 0;
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      if (appIter->second.HasData()) {
        result = 1;
      }
    }
    Leave();
    return result;
  }

  void SetData(
    APMApplicationHandle app,
    const void *buffer,
    unsigned int size)
  {
    Enter();
    CAPMWebApplicationMap::iterator appIter = m_applications.find(app);
    if (appIter != m_applications.end()) {
      CAPMWebApplication &apmapp = appIter->second;
      apmapp.SetData(buffer, size);
      apmapp.Flush(m_modules);
    }
    Leave();
  }

};

static CAPMWebContext &GetAPMWebContext()
{
  static CAPMWebContext g_context;
  return g_context;
}

//
// APMRegisterApplication
//

APMApplicationHandle APMFUNC APMRegisterApplication(
  const char *utf8applname,
  const char *utf8applid,
  const char *utf8appltier,
  const char *utf8appltierid,
  EnumAppFilterValuesCallback filterCallback)
{
  UNREFERENCED_PARAMETER(filterCallback);
  return GetAPMWebContext().RegisterApplication(utf8applname, utf8applid, utf8appltier, utf8appltierid);
}

//
// APMUnregisterApplication
//

void APMFUNC APMUnregisterApplication(APMApplicationHandle app)
{
  GetAPMWebContext().UnregisterApplication(app);
}

//
// APMRegisterModule
//

APMModuleHandle APMFUNC APMRegisterModule(
    const char *utf8modulename)
{
  return GetAPMWebContext().RegisterModule(utf8modulename);
}

//
// APMUnregisterModule
//

void APMFUNC APMUnregisterModule(
  APMModuleHandle handle)
{
  GetAPMWebContext().UnregisterModule(handle);
}

//
// APMRegisterFilterValue
//

void APMFUNC APMRegisterFilterValue(
  APMApplicationHandle app,
  const char *value,
  const char *description)
{
  GetAPMWebContext().RegisterFilterValue(app, value, description);
}

//
// APMRegisterEvent
//

void APMFUNC APMModuleRegisterEvent(
    APMModuleHandle module,
    APMEventId eventid,
    const char *description,
    const char *parameterdescription)
{
  GetAPMWebContext().ModuleRegisterEvent(module, eventid, description, parameterdescription);
}

//
// APMGetCurrentApplication
//

APMApplicationHandle APMFUNC APMGetCurrentApplication()
{
  return 0;
}

//
// APMCreateContext
//

APMTransactionHandle APMFUNC APMTxCreateContext(APMTransactionHandle thread, APMApplicationHandle app, const char *filtervalue)
{
  return GetAPMWebContext().CreateContext(thread, app, filtervalue);
}

void APMFUNC APMCreateContext(APMApplicationHandle app, const char *filtervalue)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.SetThread(ctx.CreateContext(ctx.GetThread(), app, filtervalue));
}

//
// APMGetContext
//

int APMFUNC APMTxGetContext(APMTransactionHandle thread, APMContext *context)
{
  return GetAPMWebContext().GetContext(thread, context);
}

int APMFUNC APMGetContext(APMContext *context)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  return ctx.GetContext(ctx.GetThread(), context);
}
//
// APMAttachContext
//

APMTransactionHandle APMFUNC APMTxAttachContext(
  APMTransactionHandle thread,
  APMApplicationHandle app,
  const APMContext *context)
{
  return GetAPMWebContext().AttachContext(thread, app, context);
}

void APMFUNC APMAttachContext(
  APMApplicationHandle app,
  const APMContext *context)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.SetThread(ctx.AttachContext(ctx.GetThread(), app, context));
}

//
// APMReleaseContext
//

void APMFUNC APMTxReleaseContext(APMTransactionHandle thread)
{
  GetAPMWebContext().ReleaseContext(thread);
}

void APMFUNC APMReleaseContext()
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.ReleaseContext(ctx.GetThread());
  ctx.ClearThread();
}

//
// APMGetSyncMark
//

int APMFUNC APMTxGetSyncMark(APMTransactionHandle thread, APMSyncMark *syncMark)
{
  return GetAPMWebContext().GetSyncMark(thread, syncMark);
}

int APMFUNC APMGetSyncMark(APMSyncMark *syncMark)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  return ctx.GetSyncMark(ctx.GetThread(), syncMark);
}

//
// APMSetSyncMark
//

void APMFUNC APMTxSetSyncMark(APMTransactionHandle thread, const APMSyncMark *syncMark)
{
  GetAPMWebContext().SetSyncMark(thread, syncMark);
}

void APMFUNC APMSetSyncMark(const APMSyncMark *syncMark)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.SetSyncMark(ctx.GetThread(), syncMark);
}

//
// APMHasActiveContext
//

int APMFUNC APMTxHasActiveContext(
  APMTransactionHandle thread,
  APMEventLevel level)
{
  return GetAPMWebContext().HasActiveContext(thread, level);
}

int APMFUNC APMHasActiveContext(
  APMEventLevel level)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  return ctx.HasActiveContext(ctx.GetThread(), level);
}

//
// APMReport
//

void APMFUNC APMTxReport(
  APMTransactionHandle thread,
  APMApplicationHandle app,
  const char *filtervalue,
  const char *reportkey,
  const char *description)
{
  GetAPMWebContext().Report(thread, app, filtervalue, reportkey, description);
}

void APMFUNC APMReport(
  APMApplicationHandle app,
  const char *filtervalue,
  const char *reportkey,
  const char *description)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.Report(ctx.GetThread(), app, filtervalue, reportkey, description);
}

//
// APMReportValue
//

void APMFUNC APMTxReportValue(
  APMTransactionHandle thread,
  APMApplicationHandle app,
  const char *reportkey,
  const char *key,
  const char *value)
{
  GetAPMWebContext().ReportValue(thread, app, reportkey, key, value);
}

void APMFUNC APMReportValue(
  APMApplicationHandle app,
  const char *reportkey,
  const char *key,
  const char *value)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.ReportValue(ctx.GetThread(), app, reportkey, key, value);
}

//
// APMReportContent
//

void APMFUNC APMReportContent(
  APMApplicationHandle app,
  const char *reportkey,
  const char *filename,
  const char *mimetype,
  const void *data,
  UInt64 size)
{
  UNREFERENCED_PARAMETER(app);
  UNREFERENCED_PARAMETER(reportkey);
  UNREFERENCED_PARAMETER(filename);
  UNREFERENCED_PARAMETER(mimetype);
  UNREFERENCED_PARAMETER(data);
  UNREFERENCED_PARAMETER(size);
  // NOT SUPPORTED
}

//
// APMReportFile
//

void APMFUNC APMReportFile(
  APMApplicationHandle app,
  const char *reportkey,
  const char *filename,
  const char *mimetype,
  FILE *filehandle)
{
  UNREFERENCED_PARAMETER(app);
  UNREFERENCED_PARAMETER(reportkey);
  UNREFERENCED_PARAMETER(filename);
  UNREFERENCED_PARAMETER(mimetype);
  UNREFERENCED_PARAMETER(filehandle);
  // NOT SUPPORTED
}

//
// APMIsConnected
//

int APMFUNC APMIsConnected()
{
  return 1;
}

//
// APMEvent
//

void APMFUNC APMTxEvent(
  APMTransactionHandle thread,
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags)
{
  GetAPMWebContext().Event(thread, module, eventid, level, flags);
}

void APMFUNC APMEvent(
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.Event(ctx.GetThread(), module, eventid, level, flags);
}

//
// APMEventStr
//

void APMFUNC APMTxEventStr(
  APMTransactionHandle thread,
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags,
  const char *param)
{
  GetAPMWebContext().EventStr(thread, module, eventid, level, flags, param);
}

void APMFUNC APMEventStr(
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags,
  const char *param)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.EventStr(ctx.GetThread(), module, eventid, level, flags, param);
}

//
// APMParamInit
//

APMParamHandle APMFUNC APMParamInit()
{
  return (void*)new std::stringstream();
}

//
// APMParamAdd32
//

void APMFUNC APMParamAdd32(APMParamHandle param, UInt32 value)
{
  if (param) {
    std::stringstream *ptr = (std::stringstream *)param;
    WriteItemParamNumeric(*ptr, value);
  }
}

//
// APMParamAdd64
//

void APMFUNC APMParamAdd64(APMParamHandle param, UInt64 value)
{
  if (param) {
    std::stringstream *ptr = (std::stringstream *)param;
    WriteItemParamNumeric(*ptr, value);
  }
}

//
// APMParamAddString
//

void APMFUNC APMParamAddString(APMParamHandle param, const char *value)
{
  if (param) {
    std::stringstream *ptr = (std::stringstream *)param;
    WriteItemParamString(*ptr, value);
  }
}

//
// APMParamFree
//

void APMFUNC APMParamFree(APMParamHandle param)
{
  if (param) {
    delete((std::stringstream *)param);
  }
}

//
// APMParamEvent
//

void APMFUNC APMTxEventParam(
  APMTransactionHandle thread,
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags,
  APMParamHandle param)
{
  GetAPMWebContext().EventParam(thread, module, eventid, level, flags, (std::stringstream*)param);
}

void APMFUNC APMEventParam(
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags,
  APMParamHandle param)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.EventParam(ctx.GetThread(), module, eventid, level, flags, (std::stringstream*)param);
}

void APMFUNC APMEventArgs(
  APMModuleHandle module,
  APMEventId eventid,
  APMEventLevel level,
  APMEventFlags flags,
  UInt8 paramCount,
  APMParameterType *paramTypes,
  void *paramValues,
  unsigned paramValueSize)
{
  APMParamHandle param = APMParamInit();
  if (paramCount && paramTypes && paramValues && paramValueSize) {
    char *ptr = (char*)paramValues;
    for (UInt8 i = 0; i < paramCount && (ptr - (char*)paramValues) < paramValueSize; ++i) {
      switch(paramTypes[i]) {
        case APM_PARAMETERTYPE_INT32:
          APMParamAdd32(param, *((UInt32*)ptr));
          ptr += sizeof(UInt32);
          break;
        case APM_PARAMETERTYPE_INT64:
          APMParamAdd64(param, *((UInt64*)ptr));
          ptr += sizeof(UInt64);
          break;
        case APM_PARAMETERTYPE_STRING:
          APMStringLength len = *((APMStringLength*)ptr);
          std::string value;
          if (len <= 0x8000) { // SWT_MAX_PARAMETER_STRING_LENGTH
            value.assign((char*)(ptr + sizeof(APMStringLength)), len);
          }
          APMParamAddString(param, value.c_str());
          break;
      }
    }
  }
  APMEventParam(module, eventid, level, flags, param);
  APMParamFree(param);
}

//
// APMIsCompatible
//

int APMFUNC APMIsCompatible(UInt32 size, char *message)
{
  UNREFERENCED_PARAMETER(size);
  UNREFERENCED_PARAMETER(message);
  // NOT SUPPORTED
  return TRUE;
}

int APMFUNC APMDetachThread(
  APMTransactionHandle *transaction)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  *transaction = (int)ctx.GetThread();
  return (*transaction) ? 1 : 0;
}

void APMFUNC APMAttachThread(APMTransactionHandle transaction)
{
  CAPMWebContext &ctx = GetAPMWebContext();
  ctx.SetThread(transaction);
}

int APMFUNC APMWebGetData(
  APMApplicationHandle app,
  void *buffer,
  unsigned int *size)
{
  return GetAPMWebContext().GetData(app, buffer, size);
}

void APMFUNC APMWebSetData(
  APMApplicationHandle app,
  const void *buffer,
  unsigned int size)
{
  GetAPMWebContext().SetData(app, buffer, size);
}

int APMFUNC APMWebHasData(
  APMApplicationHandle app)
{
  return GetAPMWebContext().HasData(app);
}

APMWebRecoverApplicationHandle APMFUNC APMWebRecoverApplication(void *appdata, size_t appdatasize, void *telemetrydata, size_t telemetrydatasize)
{
  CAPMWebRecoveryApplication *app = new CAPMWebRecoveryApplication();
  if (app && !app->InitRecovery(appdata, appdatasize, telemetrydata, telemetrydatasize)) {
    delete(app);
    app = NULL;
  }
  return app;
}

void APMFUNC APMWebRecoverClose(APMWebRecoverApplicationHandle app)
{
  if (app) {
    delete(static_cast<CAPMWebRecoveryApplication*>(app));
  }
}

int APMFUNC APMWebRecoverGetData(APMWebRecoverApplicationHandle app, void *buffer, unsigned int *size)
{
  if (app && static_cast<CAPMWebRecoveryApplication*>(app)->HasData()) {
    return static_cast<CAPMWebRecoveryApplication*>(app)->GetData(buffer, size);
  }
  return 0;
}

void APMFUNC APMWebRecoverConfirmData(APMWebRecoverApplicationHandle app)
{
  if (app) {
    static_cast<CAPMWebRecoveryApplication*>(app)->ConfirmData();
  }
}

void APMFUNC APMWebSetPersistHandler(struct IAPMWebPersistStore *handler)
{
  GetAPMWebContext().SetPersistHandler(handler);
}

void APMFUNC APMWebFlush()
{
  GetAPMWebContext().Flush();
}
