/*==========================================================================
  PRODUCT_VERSION_STR "15.3.7.24531"
  --------------------------------------------------------------------------
  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.

  All rights reserved. All hardware and software names used are registered
  trade names and/or registered trademarks of the respective manufacturers.

  The user of this computer program acknowledges that the above copyright
  notice, which constitutes the Universal Copyright Convention, will be
  attached at the position in the function of the computer program which the
  author has deemed to sufficiently express the reservation of copyright.
  It is prohibited for customers, users and/or third parties to remove,
  modify or move this copyright notice.
  ==========================================================================*/

#define APM_WEB_API
#define APM_WEB_API_PERSIST

#include "softwaretelemetry.h"
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>

#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <set>

#define APM_FILE_APPLICATION_EXTENSION ".apm"
#define APM_FILE_EVENT_EXTENSION      ".evt"

static std::string g_apmpersistbasedirectory;
static UInt32      g_apmpersistsize = 0;
static bool        g_apmpersistcloserecovery = false;

APMWebPersistStorageHandle APMFUNC LINUXAPMWebPersistCreateFunction(void **telemetrydatabuffer, size_t *bufferlength);
void APMFUNC LINUXAPMWebPersistWriteApplicationDataFunction(APMWebPersistStorageHandle handle, void *data, size_t length);
void APMFUNC LINUXAPMWebPersistFlushDataFunction(APMWebPersistStorageHandle handle);
void APMFUNC LINUXAPMWebPersistCloseFunction(APMWebPersistStorageHandle handle, int remove);

struct IAPMWebPersistStore g_apmpersisthandler = {
  LINUXAPMWebPersistCreateFunction,
  LINUXAPMWebPersistWriteApplicationDataFunction,
  LINUXAPMWebPersistFlushDataFunction,
  LINUXAPMWebPersistCloseFunction
};

void LINUXAPMWebPersistInit(const char *directory, UInt32 maxsize)
{
  g_apmpersistbasedirectory = directory;
  if (!g_apmpersistbasedirectory.empty()) {
    if (g_apmpersistbasedirectory.at(g_apmpersistbasedirectory.length() - 1) != '/') {
      g_apmpersistbasedirectory += '/';
    }
  }
  g_apmpersistcloserecovery = true;
  g_apmpersistsize = maxsize;
  APMWebSetPersistHandler(&g_apmpersisthandler);
}

class LINUXAPMWebPersistStorageHandler
{
private:
  std::string m_directory;
  std::string m_filebase;
  std::string m_apmFileName;
  std::string m_evtFileName;
  int m_apmFile;
  int m_evtFile;
  void *m_evtFileView;
  size_t m_evtFileSize;
  bool m_deleteFiles;

public:
  LINUXAPMWebPersistStorageHandler() :
    m_apmFile(-1), m_evtFile(-1),
    m_evtFileView(NULL), m_evtFileSize(0),
    m_deleteFiles(false)
  {
  }

  ~LINUXAPMWebPersistStorageHandler()
  {
    if (m_evtFileView) {
      munmap(m_evtFileView, m_evtFileSize);
      m_evtFileView = NULL;
      m_evtFileSize = 0;
    }
    if (m_evtFile != -1) {
      close(m_evtFile);
    }
    if (m_apmFile != -1) {
      flock(m_apmFile, LOCK_UN | LOCK_NB);
      close(m_apmFile);
    }
    if (m_deleteFiles) {
      unlink(m_apmFileName.c_str());
      unlink(m_evtFileName.c_str());
    }
  }

  void GetEventData(void **buffer, size_t *size)
  {
    if (m_evtFileView && m_evtFileSize) {
      *buffer = m_evtFileView;
      *size = m_evtFileSize;
    }
  }

  bool Create(const std::string &directory, size_t bufferSize)
  {
    bool success = false;
    m_directory = directory;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    UInt32 hash = (UInt32)tv.tv_usec;
    bool exists;
    do {
      exists = false;
      std::stringstream param;
      param.fill('0');
      param << std::setbase(16) << hash;
      m_filebase = param.str();
      m_apmFileName = m_directory + m_filebase + APM_FILE_APPLICATION_EXTENSION;
      m_evtFileName = m_directory + m_filebase + APM_FILE_EVENT_EXTENSION;

      m_apmFile = open(m_apmFileName.c_str(), O_CREAT | O_EXCL | O_RDWR, 0660);
      if (m_apmFile != -1) {
        if (0 == flock(m_apmFile, LOCK_EX | LOCK_NB)) {
          m_evtFile = open(m_evtFileName.c_str(), O_CREAT | O_EXCL | O_RDWR, 0660);
          if (m_evtFile != -1) {
            char c = 0;
            lseek(m_evtFile, bufferSize - 1, SEEK_SET);
            write(m_evtFile, &c, 1);
            m_evtFileSize = bufferSize;
            m_evtFileView = mmap(0, m_evtFileSize, PROT_READ | PROT_WRITE, MAP_SHARED, m_evtFile, 0);
            if (m_evtFileView != MAP_FAILED) {
              success = true;
            }
            else {
              m_evtFileView = NULL;
              m_evtFileSize = 0;
              close(m_evtFile);
              m_evtFile = -1;
              flock(m_apmFile, LOCK_UN);
              close(m_apmFile);
              m_apmFile = -1;
            }
          }
          else {
            if (errno == EEXIST) {
              flock(m_apmFile, LOCK_UN);
              close(m_apmFile);
              m_apmFile = -1;
              exists = true;
            }
            else {
              break;
            }
          }
        }
        else {
          close(m_apmFile);
          m_apmFile = -1;
          break;
        }
      }
      else {
        if (errno == EEXIST) {
          exists = true;
        }
        else {
          break;
        }
      }
      if (exists) {
        hash++;
      }
    } while(exists);
    return success;
  }

  void WriteApplicationData(void *data, size_t length)
  {
    if (m_apmFile != -1) {
      lseek(m_apmFile, 0, SEEK_SET);
      if (length == (size_t)write(m_apmFile, data, length)) {
        ftruncate(m_apmFile, length);
      }
    }
  }

  void FlushData()
  {
    if (m_evtFileView) {
      msync(m_evtFileView, m_evtFileSize,  MS_SYNC);
    }
  }

  void Close(bool remove)
  {
    m_deleteFiles = remove;
    delete this;
  }

};

APMWebPersistStorageHandle APMFUNC LINUXAPMWebPersistCreateFunction(void **telemetrydatabuffer, size_t *buffersize)
{
  LINUXAPMWebPersistStorageHandler *handler = new LINUXAPMWebPersistStorageHandler();
  if (handler) {
    if (handler->Create(g_apmpersistbasedirectory, g_apmpersistsize)) {
      handler->GetEventData(telemetrydatabuffer, buffersize);
    }
    else {
      delete(handler);
      handler = NULL;
    }
  }
  return handler;
}

void APMFUNC LINUXAPMWebPersistWriteApplicationDataFunction(APMWebPersistStorageHandle handle, void *data, size_t length)
{
  if (handle) {
    static_cast<LINUXAPMWebPersistStorageHandler*>(handle)->WriteApplicationData(data, length);
  }
}

void APMFUNC LINUXAPMWebPersistFlushDataFunction(APMWebPersistStorageHandle handle)
{
  if (handle) {
    static_cast<LINUXAPMWebPersistStorageHandler*>(handle)->FlushData();
  }
}

void APMFUNC LINUXAPMWebPersistCloseFunction(APMWebPersistStorageHandle handle, int remove)
{
  if (handle) {
    static_cast<LINUXAPMWebPersistStorageHandler*>(handle)->Close(remove != 0);
  }
}

static APMWebRecoverApplicationHandle g_recoveryApp = NULL;
static std::set<std::string> g_skipFileNames;

static std::string g_appFileName;
static int    g_appFile   = -1;
static void * g_appFileBuffer   = NULL;
static size_t g_appFileSize     = 0;

static std::string g_evtFileName;
static int    g_evtFile   = -1;
static void * g_evtFileBuffer   = NULL;
static size_t g_evtFileSize     = 0;

static void RecoverCloseApplication(bool deleteFiles)
{
  if (g_recoveryApp) {
    APMWebRecoverClose(g_recoveryApp);
    g_recoveryApp = NULL;
  }
  if (g_appFileBuffer) {
    free(g_appFileBuffer);
    g_appFileBuffer = NULL;
  }
  if (g_appFile != -1) {
    flock(g_appFile, LOCK_UN);
    close(g_appFile);
    g_appFile = -1;
  }
  if (g_evtFileBuffer) {
    munmap(g_evtFileBuffer, g_evtFileSize);
    g_evtFileBuffer = NULL;
    g_evtFileSize = 0;
  }
  if (g_evtFile != -1) {
    close(g_evtFile);
    g_evtFile = -1;
  }
  if (deleteFiles) {
    if (!g_appFileName.empty()) {
      unlink(g_appFileName.c_str());
    }
    if (!g_evtFileName.empty()) {
      unlink(g_evtFileName.c_str());
    }
  }
  g_appFileName.clear();
  g_evtFileName.clear();
}

static bool RecoverReadApplicationFile()
{
  bool success = false;
  g_appFile = open(g_appFileName.c_str(), O_RDONLY);
  if (g_appFile != -1) {
    if (0 == flock(g_appFile, LOCK_EX | LOCK_NB)) {
      struct stat apmfilestat;
      if (0 == fstat(g_appFile, &apmfilestat)) {
        g_appFileBuffer = malloc(apmfilestat.st_size);
        if (g_appFileBuffer) {
          if (apmfilestat.st_size == read(g_appFile, g_appFileBuffer, apmfilestat.st_size)) {
            g_appFileSize = apmfilestat.st_size;
            success = true;
          }
        }
      }
    }
  }
  if (!success) {
    if (g_appFileBuffer) {
      free(g_appFileBuffer);
      g_appFileBuffer = NULL;
    }
    if (g_appFile != -1) {
      flock(g_appFile, LOCK_UN | LOCK_NB);
      close(g_appFile);
      g_appFile = -1;
    }
  }
  return success;
}

static bool RecoverMapEventFile()
{
  bool success = false;

  g_evtFile = open(g_evtFileName.c_str(), O_RDWR);
  if (g_evtFile != -1) {
    struct stat evtfilestat;
    if (0 == fstat(g_evtFile, &evtfilestat)) {
      g_evtFileSize = evtfilestat.st_size;
      g_evtFileBuffer = mmap(0, g_evtFileSize, PROT_READ | PROT_WRITE, MAP_SHARED, g_evtFile, 0);
      if (g_evtFileBuffer != MAP_FAILED) {
        success = true;
      }
      else {
        g_evtFileBuffer = NULL;
        close(g_evtFile);
        g_evtFile = -1;
      }
    }
    else {
      close(g_evtFile);
      g_evtFile = -1;
    }
  }
  return success;
}

bool APMFUNC APMWebRecoverImpGetApplicationData(
  void *buffer,
  unsigned int *size)
{
  bool hasData = false;
  std::vector<std::string> fileNames;
  if (g_apmpersistcloserecovery) {
    g_apmpersistcloserecovery = false;
    if (g_recoveryApp) {
      RecoverCloseApplication(false);
    }
  }
  if (g_recoveryApp) {
    hasData = APMWebRecoverGetData(g_recoveryApp, buffer, size) != 0;
    if (hasData) {
      return true;
    }
    else {
      RecoverCloseApplication(true);
    }
  }

  DIR *dirStream = opendir(g_apmpersistbasedirectory.c_str());
  if (dirStream) {
    size_t len = offsetof(struct dirent, d_name) + fpathconf(dirfd(dirStream), _PC_NAME_MAX) + 1;
    struct dirent *entryp = (struct dirent *)malloc(len);
    if (entryp) {
      struct dirent *fp = NULL;
      for (;;) {
        if (readdir_r(dirStream, entryp, &fp)) {
          break;
        }
        if (fp == NULL) {
          break;
        }

        // remove files with invalid mode
        struct stat buf;
        std::string fname(g_apmpersistbasedirectory);
        fname += fp->d_name;
        if (0 == stat(fname.c_str(), &buf)) {
          if ((buf.st_mode & 0777) == 0140) {
            unlink(fname.c_str());
            continue;
          }
        }

        if (fname.length() > 4 && fname.substr(fname.length() - 4) == ".apm") {
          fileNames.push_back(fp->d_name);
        }
      }
      free(entryp);
    }
    closedir(dirStream);
  }
  std::vector<std::string>::iterator fileIter;
  for (fileIter = fileNames.begin(); !hasData && fileIter != fileNames.end(); ++fileIter) {
    std::string &filename(*fileIter);
    if (g_skipFileNames.find(filename) == g_skipFileNames.end()) {
      g_skipFileNames.insert(filename);
      g_appFileName = g_apmpersistbasedirectory + filename;
      g_evtFileName = g_apmpersistbasedirectory + filename.substr(0, filename.length() - 4) + ".evt";
      if (RecoverReadApplicationFile()) {
        bool deleteFiles = false;
        if (RecoverMapEventFile()) {
          g_recoveryApp = APMWebRecoverApplication(g_appFileBuffer, g_appFileSize, g_evtFileBuffer, g_evtFileSize);
          if (g_recoveryApp) {
            hasData = APMWebRecoverGetData(g_recoveryApp, buffer, size) != 0;
            if (hasData) {
              return 1;
            }
            else {
              deleteFiles = true;
            }
          }
        }
        RecoverCloseApplication(deleteFiles);
      }
    }
  }
  return 0;
}

void APMFUNC APMWebRecoverImpConfirmData()
{
  APMWebRecoverConfirmData(g_recoveryApp);
}

