/*==========================================================================
  PRODUCT_VERSION_STR "15.3.7.24531"
  --------------------------------------------------------------------------
  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.

  All rights reserved. All hardware and software names used are registered
  trade names and/or registered trademarks of the respective manufacturers.

  The user of this computer program acknowledges that the above copyright
  notice, which constitutes the Universal Copyright Convention, will be
  attached at the position in the function of the computer program which the
  author has deemed to sufficiently express the reservation of copyright.
  It is prohibited for customers, users and/or third parties to remove,
  modify or move this copyright notice.
  ==========================================================================*/

#define APM_WEB_API
#define APM_WEB_API_PERSIST

#include "softwaretelemetry.h"
#include <Windows.h>

#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <set>

#define APM_FILE_APPLICATION_EXTENSION ".apm"
#define APM_FILE_EVENT_EXTENSION      ".evt"

static std::string g_apmpersistbasedirectory;
static UInt32      g_apmpersistsize = 0;
static bool        g_apmpersistcloserecovery = false;

APMWebPersistStorageHandle APMFUNC WINAPMWebPersistCreateFunction(void **telemetrydatabuffer, size_t *bufferlength);
void APMFUNC WINAPMWebPersistWriteApplicationDataFunction(APMWebPersistStorageHandle handle, void *data, size_t length);
void APMFUNC WINAPMWebPersistFlushDataFunction(APMWebPersistStorageHandle handle);
void APMFUNC WINAPMWebPersistCloseFunction(APMWebPersistStorageHandle handle, int remove);

struct IAPMWebPersistStore g_apmpersisthandler = {
  WINAPMWebPersistCreateFunction,
  WINAPMWebPersistWriteApplicationDataFunction,
  WINAPMWebPersistFlushDataFunction,
  WINAPMWebPersistCloseFunction
};

void WINAPMWebPersistInit(const char *directory, UInt32 maxsize)
{
  g_apmpersistbasedirectory = directory;
  if (!g_apmpersistbasedirectory.empty()) {
    if (g_apmpersistbasedirectory.at(g_apmpersistbasedirectory.length() - 1) != '\\') {
      g_apmpersistbasedirectory += '\\';
    }
  }
  g_apmpersistcloserecovery = true;
  g_apmpersistsize = maxsize;
  APMWebSetPersistHandler(&g_apmpersisthandler);
}

class WINAPMWebPersistStorageHandler
{
private:
  std::string m_directory;
  std::string m_filebase;
  std::string m_apmFileName;
  std::string m_evtFileName;
  HANDLE m_apmFile;
  HANDLE m_evtFile;
  HANDLE m_evtFileMapping;
  LPVOID m_evtFileView;
  size_t m_evtFileSize;
  bool   m_deleteFiles;

public:
  WINAPMWebPersistStorageHandler() :
    m_apmFile(INVALID_HANDLE_VALUE), m_evtFile(INVALID_HANDLE_VALUE),
    m_evtFileMapping(NULL), m_evtFileView(NULL), m_evtFileSize(0),
    m_deleteFiles(false)
  {
  }

  ~WINAPMWebPersistStorageHandler()
  {
    if (m_evtFileView) {
      UnmapViewOfFile(m_evtFileView);
      m_evtFileView = NULL;
      m_evtFileSize = 0;
    }
    if (m_evtFileMapping != NULL) {
      CloseHandle(m_evtFileMapping);
    }
    if (m_evtFile != NULL) {
      CloseHandle(m_evtFile);
    }
    if (m_apmFile != NULL) {
      CloseHandle(m_apmFile);
    }
    if (m_deleteFiles) {
      DeleteFile(m_apmFileName.c_str());
      DeleteFile(m_evtFileName.c_str());
    }
  }

  void GetEventData(void **buffer, size_t *size)
  {
    if (m_evtFileView && m_evtFileSize) {
      *buffer = m_evtFileView;
      *size = m_evtFileSize;
    }
  }

  bool Create(const std::string &directory, size_t bufferSize)
  {
    bool success = false;
    m_directory = directory;
    LARGE_INTEGER time;
    QueryPerformanceCounter(&time);
    UInt32 hash = (UInt32)time.LowPart;
    bool exists;
    do {
      exists = false;
      std::stringstream param;
      param.fill('0');
      param << std::setbase(16) << hash;
      m_filebase = param.str();
      m_apmFileName = m_directory + m_filebase + APM_FILE_APPLICATION_EXTENSION;
      m_evtFileName = m_directory + m_filebase + APM_FILE_EVENT_EXTENSION;
      m_apmFile = CreateFile(m_apmFileName.c_str(), GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
      if (m_apmFile != INVALID_HANDLE_VALUE) {
        m_evtFile = CreateFile(m_evtFileName.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
        if (m_evtFile != INVALID_HANDLE_VALUE) {
          DWORD dwFileSize = (DWORD)bufferSize;
          dwFileSize = dwFileSize + 0xfff & ~0xfff;
          m_evtFileMapping = CreateFileMapping(m_evtFile, NULL, PAGE_READWRITE, 0, dwFileSize, NULL);
          if (m_evtFileMapping != NULL) {
            m_evtFileView = MapViewOfFile(m_evtFileMapping, FILE_MAP_WRITE, 0, 0, dwFileSize);
            if (m_evtFileView != NULL) {
              m_evtFileSize = dwFileSize;
              success = true;
            }
            else {
              CloseHandle(m_evtFileMapping);
              m_evtFileMapping = NULL;
              CloseHandle(m_evtFile);
              m_evtFile = INVALID_HANDLE_VALUE;
              CloseHandle(m_apmFile);
              m_apmFile = INVALID_HANDLE_VALUE;
            }
          }
          else {
            CloseHandle(m_evtFile);
            m_evtFile = INVALID_HANDLE_VALUE;
            CloseHandle(m_apmFile);
            m_apmFile = INVALID_HANDLE_VALUE;
          }
        }
        else {
          CloseHandle(m_apmFile);
          m_apmFile = INVALID_HANDLE_VALUE;
          exists = true;
        }
      }
      else {
        if (GetLastError() == ERROR_FILE_EXISTS) {
          exists = true;
        }
        else {
          break;
        }
      }
      if (exists) {
        hash = 16777619U * hash + (UInt32)GetCurrentThreadId() % 47 + 1;
      }
    } while(exists);
    return success;
  }

  void WriteApplicationData(void *data, size_t length)
  {
    if (m_apmFile != INVALID_HANDLE_VALUE) {
      LARGE_INTEGER move;
      move.QuadPart = 0;
      SetFilePointerEx(m_apmFile, move, NULL, FILE_BEGIN);
      DWORD written = 0;
      if (WriteFile(m_apmFile, data, (DWORD)length, &written, NULL)) {
        SetEndOfFile(m_apmFile);
      }
    }
  }

  void FlushData()
  {
    if (m_evtFileView) {
      FlushViewOfFile(m_evtFileView, 0);
    }
  }

  void Close(bool remove)
  {
    m_deleteFiles = remove;
    delete this;
  }

};

APMWebPersistStorageHandle APMFUNC WINAPMWebPersistCreateFunction(void **telemetrydatabuffer, size_t *buffersize)
{
  WINAPMWebPersistStorageHandler *handler = new WINAPMWebPersistStorageHandler();
  if (handler) {
    if (handler->Create(g_apmpersistbasedirectory, g_apmpersistsize)) {
      handler->GetEventData(telemetrydatabuffer, buffersize);
    }
    else {
      delete(handler);
      handler = NULL;
    }
  }
  return handler;
}

void APMFUNC WINAPMWebPersistWriteApplicationDataFunction(APMWebPersistStorageHandle handle, void *data, size_t length)
{
  if (handle) {
    static_cast<WINAPMWebPersistStorageHandler*>(handle)->WriteApplicationData(data, length);
  }
}

void APMFUNC WINAPMWebPersistFlushDataFunction(APMWebPersistStorageHandle handle)
{
  if (handle) {
    static_cast<WINAPMWebPersistStorageHandler*>(handle)->FlushData();
  }
}

void APMFUNC WINAPMWebPersistCloseFunction(APMWebPersistStorageHandle handle, int remove)
{
  if (handle) {
    static_cast<WINAPMWebPersistStorageHandler*>(handle)->Close(remove != 0);
  }
}

static APMWebRecoverApplicationHandle g_recoveryApp = NULL;
static std::set<std::string> g_skipFileNames;

static std::string g_appFileName;
static HANDLE g_appFileHandle   = INVALID_HANDLE_VALUE;
static void * g_appFileBuffer   = NULL;
static size_t g_appFileSize     = 0;

static std::string g_evtFileName;
static HANDLE g_evtFile   = INVALID_HANDLE_VALUE;
static HANDLE g_evtFileMapping  = INVALID_HANDLE_VALUE;
static void * g_evtFileBuffer   = NULL;
static size_t g_evtFileSize     = 0;

static void RecoverCloseApplication(bool deleteFiles)
{
  if (g_recoveryApp) {
    APMWebRecoverClose(g_recoveryApp);
    g_recoveryApp = NULL;
  }
  if (g_appFileBuffer) {
    free(g_appFileBuffer);
    g_appFileBuffer = NULL;
  }
  if (g_appFileHandle != INVALID_HANDLE_VALUE) {
    CloseHandle(g_appFileHandle);
    g_appFileHandle = INVALID_HANDLE_VALUE;
  }
  if (g_evtFileBuffer) {
    UnmapViewOfFile(g_evtFileBuffer);
    g_evtFileBuffer = NULL;
    g_evtFileSize = 0;
  }
  if (g_evtFileMapping != NULL) {
    CloseHandle(g_evtFileMapping);
    g_evtFileMapping = NULL;
  }
  if (g_evtFile != INVALID_HANDLE_VALUE) {
    CloseHandle(g_evtFile);
    g_evtFile = INVALID_HANDLE_VALUE;
  }
  if (deleteFiles) {
    if (!g_appFileName.empty()) {
      DeleteFile(g_appFileName.c_str());
    }
    if (!g_evtFileName.empty()) {
      DeleteFile(g_evtFileName.c_str());
    }
  }
  g_appFileName.clear();
  g_evtFileName.clear();
}

static bool RecoverReadApplicationFile()
{
  bool success = false;
  g_appFileHandle = CreateFile(g_appFileName.c_str(), GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (g_appFileHandle != INVALID_HANDLE_VALUE) {
    DWORD fileSize = GetFileSize(g_appFileHandle, NULL);
    if (fileSize > 0) {
      g_appFileBuffer = malloc(fileSize);
      if (g_appFileBuffer) {
        DWORD fileRead = 0;
        if (ReadFile(g_appFileHandle, g_appFileBuffer, fileSize, &fileRead, NULL)) {
          g_appFileSize = fileRead;
          success = true;
        }
      }
    }
  }
  if (!success) {
    if (g_appFileBuffer) {
      free(g_appFileBuffer);
      g_appFileBuffer = NULL;
    }
    if (g_appFileHandle != INVALID_HANDLE_VALUE) {
      CloseHandle(g_appFileHandle);
      g_appFileHandle = INVALID_HANDLE_VALUE;
    }
  }
  return success;
}

static bool RecoverMapEventFile()
{
  bool success = false;
  g_evtFile = CreateFile(g_evtFileName.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (g_evtFile != INVALID_HANDLE_VALUE) {
    DWORD fileSize = GetFileSize(g_evtFile, NULL);
    if (fileSize > 0) {
      g_evtFileMapping = CreateFileMapping(g_evtFile, NULL, PAGE_READWRITE, 0, fileSize, NULL);
      if (g_evtFileMapping != NULL) {
        g_evtFileBuffer = MapViewOfFile(g_evtFileMapping, FILE_MAP_WRITE, 0, 0, fileSize);
        if (g_evtFileBuffer != NULL) {
          g_evtFileSize = fileSize;
          success = true;
        }
        else {
          CloseHandle(g_evtFileMapping);
          g_evtFileMapping = NULL;
          CloseHandle(g_evtFile);
          g_evtFile = INVALID_HANDLE_VALUE;
        }
      }
      else {
        CloseHandle(g_evtFile);
        g_evtFile = INVALID_HANDLE_VALUE;
      }
    }
    CloseHandle(g_evtFile);
    g_evtFile = INVALID_HANDLE_VALUE;
  }
  return success;
}

bool APMFUNC APMWebRecoverImpGetApplicationData(
  void *buffer,
  unsigned int *size)
{
  bool hasData = false;
  std::vector<std::string> fileNames;
  if (g_apmpersistcloserecovery) {
    g_apmpersistcloserecovery = false;
    if (g_recoveryApp) {
      RecoverCloseApplication(false);
    }
  }
  if (g_recoveryApp) {
    hasData = APMWebRecoverGetData(g_recoveryApp, buffer, size) != 0;
    if (hasData) {
      return true;
    }
    else {
      RecoverCloseApplication(true);
    }
  }

  WIN32_FIND_DATA FindFileData;
  HANDLE hFind = FindFirstFile((g_apmpersistbasedirectory + "*.apm").c_str(), &FindFileData);
  if (hFind != INVALID_HANDLE_VALUE) {
    do {
      if ((FindFileData.dwFileAttributes &
            (FILE_ATTRIBUTE_DEVICE |
            FILE_ATTRIBUTE_DIRECTORY |
            FILE_ATTRIBUTE_OFFLINE |
            FILE_ATTRIBUTE_READONLY |
            FILE_ATTRIBUTE_SYSTEM)) == 0)
      {
        fileNames.push_back(FindFileData.cFileName);
      }
    }
    while (FindNextFile(hFind, &FindFileData));
    FindClose(hFind);
  }
  std::vector<std::string>::iterator fileIter;
  for (fileIter = fileNames.begin(); !hasData && fileIter != fileNames.end(); ++fileIter) {
    std::string &filename(*fileIter);
    if (filename.length() > 4 && filename.substr(filename.length() - 4) == ".apm") {
      if (g_skipFileNames.find(filename) == g_skipFileNames.end()) {
        g_skipFileNames.insert(filename);
        g_appFileName = g_apmpersistbasedirectory + filename;
        g_evtFileName = g_apmpersistbasedirectory + filename.substr(0, filename.length() - 4) + ".evt";
        if (RecoverReadApplicationFile()) {
          bool deleteFiles = false;
          if (RecoverMapEventFile()) {
            g_recoveryApp = APMWebRecoverApplication(g_appFileBuffer, g_appFileSize, g_evtFileBuffer, g_evtFileSize);
            if (g_recoveryApp) {
              hasData = APMWebRecoverGetData(g_recoveryApp, buffer, size) != 0;
              if (hasData) {
                return 1;
              }
              else {
                deleteFiles = true;
              }
            }
          }
          RecoverCloseApplication(deleteFiles);
        }
      }
    }
  }
  return 0;
}

void APMFUNC APMWebRecoverImpConfirmData()
{
  APMWebRecoverConfirmData(g_recoveryApp);
}
