/*==========================================================================
  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.

  All rights reserved. All hardware and software names used are registered
  trade names and/or registered trademarks of the respective manufacturers.

  The user of this computer program acknowledges that the above copyright
  notice, which constitutes the Universal Copyright Convention, will be
  attached at the position in the function of the computer program which the
  author has deemed to sufficiently express the reservation of copyright.
  It is prohibited for customers, users and/or third parties to remove,
  modify or move this copyright notice.
  ==========================================================================*/
package com.apptelemetry.apm.client.feedback;

import java.net.MalformedURLException;
import java.net.URL;

import android.app.Activity;
import android.content.Intent;
import android.os.Bundle;
import android.view.View;
import android.widget.EditText;

import com.apptelemetry.apm.client.APM;
import com.apptelemetry.apm.client.APMModule;
import com.apptelemetry.apm.client.APMReportContent;

public class APMDemoApplication extends Activity {
	APM apm;
	APMModule moduleEvent;
	static String currentURL;

	static {
		currentURL = "";
	}

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_apmdemo_application);
		APM.init(this.getApplication(), "");
		apm = APM.getInstance();
		apm.RegisterApplication("Fabasoft app.telemetry", "Test Application", "Android", "Test 1");
		moduleEvent = APM.getModule("Event");
		moduleEvent.RegisterEvent(1001, "sendFeedback", "message");
		moduleEvent.RegisterEvent(1002, "secondThread", "message");
	}

	@Override
	protected void onResume() {
		super.onResume();
		EditText editText = (EditText) findViewById(R.id.edit_url);
		if (null != editText) {
			editText.setText(currentURL);
		}
	}

	/**
	 * Called when the user clicks the Send button
	 */
	public void sendFeedback(View view) {
		APMReportContent screenshot = APM.createScreenshot(view);

		long thread = apm.CreateContextTx(0, "sendFeedback");
		Intent intent = new Intent(this, FeedbackDialogActivity.class);
		EditText editText = (EditText) findViewById(R.id.edit_message);
		moduleEvent.EventStrTx(thread, 1001, APM.EVENT_LEVEL_NORMAL, APM.FLAG_ENTER, editText.getText().toString());

		byte[] contextToken = apm.GetContextTx(thread);

		long thread2 = apm.AttachContextTx(0, contextToken);
		moduleEvent.EventStrTx(thread2, 1002, APM.EVENT_LEVEL_DETAIL, APM.FLAG_ENTER, editText.getText().toString());
		moduleEvent.EventTx(thread2, 1002, APM.EVENT_LEVEL_NORMAL, APM.FLAG_LEAVE);
		byte[] syncMark = apm.GetSyncMarkTx(thread2);
		apm.ReleaseContextTx(thread2);

		apm.SetSyncMarkTx(thread, syncMark);
		String message = editText.getText().toString();
		intent.putExtra(FeedbackDialogActivity.EXTRA_SCREENSHOT, screenshot);
		intent.putExtra(FeedbackDialogActivity.EXTRA_TOPIC, message);
		startActivity(intent);
		moduleEvent.EventTx(thread, 1001, APM.EVENT_LEVEL_NORMAL, APM.FLAG_LEAVE);
		apm.ReleaseContextTx(thread);
	}

	/**
	 * Called when the user clicks the setURL button
	 */
	public void setURL(View view) {
		EditText editText = (EditText) findViewById(R.id.edit_url);
		currentURL = editText.getText().toString();
		try {
			if (0 != currentURL.length()) {
				URL url = new URL(currentURL);
			}
			APM.setURL(currentURL);
		}
		catch (MalformedURLException e) {
			editText.setText("");
		}
	}
}
