﻿/* ==========================================================================
 * Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
 *
 * All rights reserved. All hardware and software names used are registered
 * trade names and/or registered trademarks of the respective manufacturers.
 *
 * The user of this computer program acknowledges that the above copyright
 * notice, which constitutes the Universal Copyright Convention, will be
 * attached at the position in the function of the computer program which the
 * author has deemed to sufficiently express the reservation of copyright.
 * It is prohibited for customers, users and/or third parties to remove,
 * modify or move this copyright notice.
 * ========================================================================= */

using System;
using System.Net;
using System.IO;
using Microsoft.Win32;
using apm;

public partial class WebShop : System.Web.UI.Page
{
    public static bool g_registered = false;

    public const UInt64 EVENT_PROCESS_START = 100;
    public const UInt64 EVENT_PROCESS_PARAMS = 101;
    public const UInt64 EVENT_PROCESS_END = 102;
    public const UInt64 EVENT_GET_REPOSITORY_DATA = 110;
    public const UInt64 EVENT_GET_REPOSITORY_DATA_REQUEST = 111;
    public const UInt64 EVENT_GET_STOCK_SERVICE_DATA = 120;
    public const UInt64 EVENT_GET_STOCK_SERVICE_DATA_REQUEST = 121;

    public static APMModule g_samplemodule;
    public static APMModule g_repositorymodule;
    public static APMModule g_stockmodule;
    public static APMApplication g_application;

    public int m_requestsize;
    public int m_workerthreads;
    public String m_stockserviceurl;
    public String m_repositoryserviceurl;

    public WebShop()
    {
      // register application with APM Software-Telemetry

      if (!WebShop.g_registered)
      {
        WebShop.g_application = new APMApplication("Fabasoft app.telemetry", "Demo", "Frontend Service", "Online Shop");
        WebShop.g_application.RegisterFilterValue("Article ID", "Query");
        WebShop.g_application.RegisterFilterValue("AJAX Request", "Online Check");

        // register all modules
        WebShop.g_samplemodule = new APMModule("C# Service");
        WebShop.g_samplemodule.RegisterEvent(EVENT_PROCESS_START, "Process Request Start", "");
        WebShop.g_samplemodule.RegisterEvent(EVENT_PROCESS_PARAMS, "Process Parameters", "Thread Counter;Request Size");
        WebShop.g_samplemodule.RegisterEvent(EVENT_PROCESS_END, "Process Request End", "");

        WebShop.g_repositorymodule = new APMModule("Repository Request");
        WebShop.g_repositorymodule.RegisterEvent(EVENT_GET_REPOSITORY_DATA, "Get Repository Data", "Request Size");
        WebShop.g_repositorymodule.RegisterEvent(EVENT_GET_REPOSITORY_DATA_REQUEST, "Get Repository Data", "");

        WebShop.g_stockmodule = new APMModule("Stock Request");
        WebShop.g_stockmodule.RegisterEvent(EVENT_GET_STOCK_SERVICE_DATA, "Get Stock Data", "Thread Count");
        WebShop.g_stockmodule.RegisterEvent(EVENT_GET_STOCK_SERVICE_DATA_REQUEST, "Get Stock Data HTTP Request", "");

        g_registered = true;
      }
      // read configuration parameters from registry
      m_requestsize = 1;
      m_workerthreads = 10;
      try
      {
        RegistryKey keySoftware = Registry.LocalMachine.OpenSubKey("Software");
        RegistryKey keyApplStrudl = keySoftware.OpenSubKey("Fabasoft app.telemetry");
        RegistryKey keyDemoApplication = keyApplStrudl.OpenSubKey("OnlineShopSample");
        m_requestsize = (int)keyDemoApplication.GetValue("RequestSize", 1);
        m_workerthreads = (int)keyDemoApplication.GetValue("WorkerThreads", 10);
        m_repositoryserviceurl = (String)keyDemoApplication.GetValue("RepositoryServiceUrl",
          "http://localhost:85/OnlineShopRepository/OnlineShopRepository.dll");
        m_stockserviceurl = (String)keyDemoApplication.GetValue("StockServiceUrl",
          "http://localhost:8085/");
      }
      catch
      {
      }
    }

    public void WebShopWebMethod(object sender, EventArgs e)
    {


      // start Software-Stelemetry session
      WebShop.g_application.CreateContext("Article ID");
      if (apmcontext.Value != "")
      {
        byte[] rawcontext = System.Convert.FromBase64String(apmcontext.Value);
        WebShop.g_application.AttachContext(rawcontext);
      }
      // mark start of request processing
      WebShop.g_samplemodule.Event(
        WebShop.EVENT_PROCESS_START,
        APM.EVENT_LEVEL_LOG,
        APM.FLAG_ENTER);

      APMParameter param = new APMParameter();
      param.AddInt32((UInt32)m_workerthreads);
      param.AddInt32((UInt32)m_requestsize);

      WebShop.g_samplemodule.EventParam(
        WebShop.EVENT_PROCESS_PARAMS,
        APM.EVENT_LEVEL_LOG,
        APM.FLAG_NONE,
        param);

      // get information from repository service
      GetRepositoryInfo();

      // get information from stock service
      GetStockServiceInfo();

      // mark end of request processing
      WebShop.g_samplemodule.Event(
        WebShop.EVENT_PROCESS_END,
        APM.EVENT_LEVEL_LOG,
        APM.FLAG_LEAVE);

      byte[] rawResponseSyncMark = new byte[16];
      WebShop.g_application.GetSyncMark(out rawResponseSyncMark);
      string responseSyncMark = System.Convert.ToBase64String(rawResponseSyncMark);
      apmcontext.Value = responseSyncMark;

      // end Software-Telemetry session
      WebShop.g_application.ReleaseContext();
    }

    public void GetRepositoryInfo()
    {
      // mark start of repository data processing
      WebShop.g_repositorymodule.EventStr(
        WebShop.EVENT_GET_REPOSITORY_DATA,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_ENTER,
        m_requestsize.ToString());

      int length = 0;
      String responseText = "";
      String descriptionText = "";

      do
      {
        // mark start of repository request processing
        WebShop.g_repositorymodule.Event(
          WebShop.EVENT_GET_REPOSITORY_DATA_REQUEST,
          APM.EVENT_LEVEL_DETAIL,
          APM.FLAG_ENTER);

        HttpWebRequest request = (HttpWebRequest)WebRequest.Create(m_repositoryserviceurl);
        request.Method = "POST";

        byte[] context;
        WebShop.g_application.GetContext(out context);
        request.Headers.Add("TelemetryContext", System.Convert.ToBase64String(context));

        request.ContentType = "text/plain";

        String requestvalue = box.Text + "," + length.ToString() + "," + m_requestsize.ToString();
        request.ContentLength = requestvalue.Length;
        Stream requestStream = request.GetRequestStream();
        StreamWriter requestWriter = new StreamWriter(requestStream);
        requestWriter.Write(requestvalue);
        requestWriter.Close();

        HttpWebResponse response = (HttpWebResponse)request.GetResponse();
        String syncMark = response.GetResponseHeader("TelemetrySyncMark");
        if (syncMark != null)
        {
          WebShop.g_application.SetSyncMark(Convert.FromBase64String(syncMark));
        }
        StreamReader responseReader = new StreamReader(response.GetResponseStream());

        responseText = responseReader.ReadToEnd();
        descriptionText += responseText;
        length += m_requestsize;
        responseReader.Close();
        response.Close();

        // mark end of repository request processing
        WebShop.g_repositorymodule.Event(
          WebShop.EVENT_GET_REPOSITORY_DATA_REQUEST,
          APM.EVENT_LEVEL_DETAIL,
          APM.FLAG_LEAVE);

      } while (responseText.Length >= m_requestsize);

      description.Text = descriptionText;

      // mark end of repository data processing
      WebShop.g_repositorymodule.Event(
        WebShop.EVENT_GET_REPOSITORY_DATA,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_LEAVE);
    }

    void GetStockServiceInfo()
    {
      WebShop.g_stockmodule.EventStr(
        WebShop.EVENT_GET_STOCK_SERVICE_DATA,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_ENTER,
        m_workerthreads.ToString());

      HttpWebRequest request = (HttpWebRequest)WebRequest.Create(m_stockserviceurl);
      request.Method = "POST";

      byte[] context;
      WebShop.g_application.GetContext(out context);
      string context_string = System.Convert.ToBase64String(context);
      request.Headers.Add("TelemetryContext", context_string);

      request.ContentType = "text/plain";

      WebShop.g_stockmodule.Event(
        WebShop.EVENT_GET_STOCK_SERVICE_DATA_REQUEST,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_ENTER | APM.FLAG_WAIT);

      String requestvalue = "ID=" + box.Text + ",THREADS=" + m_workerthreads.ToString();
      request.ContentLength = requestvalue.Length;
      Stream requestStream = request.GetRequestStream();
      StreamWriter requestWriter = new StreamWriter(requestStream);
      requestWriter.Write(requestvalue);
      requestWriter.Close();

      HttpWebResponse response = (HttpWebResponse)request.GetResponse();

      String syncMark = response.GetResponseHeader("TelemetrySyncMark");
      if (syncMark != null)
      {
          WebShop.g_application.SetSyncMark(Convert.FromBase64String(syncMark));
      }

      StreamReader responseReader = new StreamReader(response.GetResponseStream());

      availability.Text = responseReader.ReadToEnd();

      WebShop.g_stockmodule.Event(
        WebShop.EVENT_GET_STOCK_SERVICE_DATA_REQUEST,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_LEAVE | APM.FLAG_WAIT);

      responseReader.Close();
      response.Close();

      WebShop.g_stockmodule.Event(
        WebShop.EVENT_GET_STOCK_SERVICE_DATA,
        APM.EVENT_LEVEL_DETAIL,
        APM.FLAG_LEAVE);
    }

    public bool CheckRepositoryService()
    {
      try
      {
        HttpWebRequest request = (HttpWebRequest)WebRequest.Create(m_repositoryserviceurl);
        request.Method = "POST";

        byte[] context;
        WebShop.g_application.GetContext(out context);
        request.Headers.Add("TelemetryContext", System.Convert.ToBase64String(context));

        request.ContentType = "text/plain";

        String requestvalue = ",0,0";
        request.ContentLength = requestvalue.Length;
        Stream requestStream = request.GetRequestStream();
        StreamWriter requestWriter = new StreamWriter(requestStream);
        requestWriter.Write(requestvalue);
        requestWriter.Close();

        HttpWebResponse response = (HttpWebResponse)request.GetResponse();
        String syncMark = response.GetResponseHeader("TelemetrySyncMark");
        if (syncMark != null)
        {
          WebShop.g_application.SetSyncMark(Convert.FromBase64String(syncMark));
        }
        StreamReader responseReader = new StreamReader(response.GetResponseStream());

        string responseText = responseReader.ReadToEnd();
      }
      catch
      {
        return false;
      }
      return true;
    }

    public bool CheckStockService()
    {
      try
      {
        HttpWebRequest request = (HttpWebRequest)WebRequest.Create(m_stockserviceurl);
        request.Method = "POST";

        byte[] context;
        WebShop.g_application.GetContext(out context);
        string context_string = System.Convert.ToBase64String(context);
        request.Headers.Add("TelemetryContext", context_string);

        request.ContentType = "text/plain";

        WebShop.g_stockmodule.Event(
          WebShop.EVENT_GET_STOCK_SERVICE_DATA_REQUEST,
          APM.EVENT_LEVEL_DETAIL,
          APM.FLAG_ENTER | APM.FLAG_WAIT);

        String requestvalue = "ID=,THREADS=10";
        request.ContentLength = requestvalue.Length;
        Stream requestStream = request.GetRequestStream();
        StreamWriter requestWriter = new StreamWriter(requestStream);
        requestWriter.Write(requestvalue);
        requestWriter.Close();

        HttpWebResponse response = (HttpWebResponse)request.GetResponse();

        String syncMark = response.GetResponseHeader("TelemetrySyncMark");
        if (syncMark != null)
        {
          WebShop.g_application.SetSyncMark(Convert.FromBase64String(syncMark));
        }

        StreamReader responseReader = new StreamReader(response.GetResponseStream());

        string responseText = responseReader.ReadToEnd();
      }
      catch
      {
        return false;
      }

      return true;
    }

}
