/* ==========================================================================
 * Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
 *
 * All rights reserved. All hardware and software names used are registered
 * trade names and/or registered trademarks of the respective manufacturers.
 *
 * The user of this computer program acknowledges that the above copyright
 * notice, which constitutes the Universal Copyright Convention, will be
 * attached at the position in the function of the computer program which the
 * author has deemed to sufficiently express the reservation of copyright.
 * It is prohibited for customers, users and/or third parties to remove,
 * modify or move this copyright notice.
 * ========================================================================= */

#include <string>

class Base64Encoder {
private:
  static char EncodeChar(BYTE binary)
  {
    const char b64digits[] = TEXT("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/");
    if (binary < 64) {
      return b64digits[binary];
    }
    return '0';
  }

public:
  static std::string Convert(const BYTE *buffer, size_t length)
  {
    const unsigned char *ptr = (const unsigned char *)buffer;
    std::stringstream stream;
    while (length >= 3) {
      stream << EncodeChar(((ptr[0] & 0xFC) >> 2));
      stream << EncodeChar(((ptr[0] & 0x03) << 4) | ((ptr[1] & 0xF0) >> 4));
      stream << EncodeChar(((ptr[1] & 0x0F) << 2) | ((ptr[2] & 0xC0) >> 6));
      stream << EncodeChar(((ptr[2] & 0x3F)));
      ptr += 3;
      length -= 3;
    }
    if (length == 2) {
      stream << EncodeChar(((ptr[0] & 0xFC) >> 2));
      stream << EncodeChar(((ptr[0] & 0x03) << 4) | ((ptr[1] & 0xF0) >> 4));
      stream << EncodeChar(((ptr[1] & 0x0F) << 2));
      stream << '=';
    }
    else if (length == 1) {
      stream << EncodeChar(((ptr[0] & 0xFC) >> 2));
      stream << EncodeChar(((ptr[0] & 0x03) << 4));
      stream << '=';
      stream << '=';
    }
    return stream.str();
  }

};

class Base64Decoder {
private:
  static BYTE DecodeChar(char b64)
  {
    if (b64 >= 'A' && b64 <= 'Z') {
      return (BYTE)(b64 - 'A');
    }
    if (b64 >= 'a' && b64 <= 'z') {
      return (BYTE)(b64 - 'a' + 26);
    }
    if (b64 >= '0' && b64 <= '9') {
      return (BYTE)(b64 - '0' + 52);
    }
    if (b64 == '+') {
      return 62;
    }
    if (b64 == '/') {
      return 63;
    }
    return 64;
  }

public:
  static size_t Convert(const std::string &source, BYTE *buffer, size_t length)
  {
    size_t targetLength = 0;
    int targetByte = 0;
    const char *ptr = source.c_str();
    byte previousBin = 0;
    while (*ptr) {
      while (*ptr != 0 && (*ptr == ' ' || *ptr == '\r' || *ptr == '\n' || *ptr == '\t')) {
        ptr++;
      }
      char b = *ptr;
      if (b == 0 || b == '=') {
        break;
      }
      BYTE bin = DecodeChar(b);
      if (bin != 64) {
        switch (targetByte) {
        case 0:
          previousBin = bin;
          targetByte++;
          break;
        case 1:
          if (targetLength < length) {
            *buffer = (previousBin << 2) | ((bin >> 4) & 0x03);
            buffer++;
            targetLength++;
          }
          previousBin = bin;
          targetByte++;
          break;
        case 2:
          if (targetLength < length) {
            *buffer = ((previousBin << 4) & 0xF0) | ((bin >> 2) & 0x0F);
            buffer++;
            targetLength++;
          }
          previousBin = bin;
          targetByte++;
          break;
        case 3:
          if (targetLength < length) {
            *buffer = ((previousBin << 6) & 0xC0) | (bin & 0x3F);
            buffer++;
            targetLength++;
          }
          targetByte = 0;
          break;
        }
      }
      ptr++;
    }
    return targetLength;
  }
};
