/* ==========================================================================
 * Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
 *
 * All rights reserved. All hardware and software names used are registered
 * trade names and/or registered trademarks of the respective manufacturers.
 *
 * The user of this computer program acknowledges that the above copyright
 * notice, which constitutes the Universal Copyright Convention, will be
 * attached at the position in the function of the computer program which the
 * author has deemed to sufficiently express the reservation of copyright.
 * It is prohibited for customers, users and/or third parties to remove,
 * modify or move this copyright notice.
 * ========================================================================= */

#include <httpext.h>
#include <string>
#include <list>
#include <windows.h>
#include <ole2.h>
#include <atlbase.h>
#include <atlcom.h>
#include "webserver.h"
#include <softwaretelemetry.h>

//------------------------------------------------------------------------------
// atl
//------------------------------------------------------------------------------

CComModule _AtlModule;

BEGIN_OBJECT_MAP(ObjectMap)
END_OBJECT_MAP()


//------------------------------------------------------------------------------
// APM Software-Telemetry definitions
//------------------------------------------------------------------------------

APMApplicationHandle g_app;
APMModuleHandle g_module;
const char *g_result_text =
  "Fabasoft app.telemetry 2010 provides powerful and easy to use tools "
  "for state-of-the-art application management. As demands on IT services "
  "operations dramatically increase to provide strong availability and response "
  "time service level agreements, more than a classic system management "
  "infrastructure is required. With its tight and efficient integration into "
  "Fabasoft software products, Fabasoft app.telemetry is the unrivaled "
  "answer for applications management. The all new Software Telemetry is the "
  "best solution for tracking down and solving user-specific problems.\n"

  "Where are the performance bottlenecks in my system? This question becomes "
  "increasingly important as systems move towards service-oriented architectures "
  "(SOA). A single request utilizes different services on many servers and "
  "potentially across several sites. Finding a performance bottleneck impacting "
  "a particular user challenges operations management. Which sites processed a "
  "specific request? Which services on which servers were involved? Where are "
  "the delays coming from?\n"

  "Fabasoft app.telemetry follows the users request through your "
  "services, servers and sites. Collecting detailed information on any involved "
  "services, Fabasoft app.telemetry immediately shows the requests "
  "route through the system and making it easy for operations management to "
  "discover and identify the performance-critical locations.\n\n"

  "SOFTWARE-TELEMETRY\n"
  "A single request initiated by a user can utilize several applications, a "
  "specific site, a single web server in a web server farm, a single service on "
  "the web server, a specific cluster node of a database cluster, a single "
  "database instance on the cluster node, etc. It is particularly challenging to "
  "solve such performance problems in a complex distributed environment, since it "
  "is first necessary to identify the affected applications, services and servers.\n"

  "Fabasoft app.telemetry is designed to follow selected requests through "
  "the environment. As activation is done online, affected servers and services are "
  "immediately identified and round-trips between services are displayed "
  "graphically.\n"
  "Fabasoft app.telemetry provides unrivaled information for "
  "performance analyses and problem resolution.\n\n"

  "SERVICE ACCESS STATISTICS\n"
  "In addition to the software telemetry features designed for selected requests, "
  "Fabasoft app.telemetry also provides a complete overview of all "
  "requests processed by your system. app.telemetry service access statistics can "
  "collect any individual request processed by the service and can offer details "
  "like request time, duration within the service (response time), bytes transferred "
  "over the network or many other statistics. Uploaded and preprocessed in a data "
  "warehouse, app.telemetry service statistics are a powerful tool for performance "
  "analysis and service level management. Request duration analysis is especially "
  "useful and can provide an exhaustive presentation of the services overall response "
  "time in the data center, and giving operations management objective information "
  "about their system.\n\n"

  "SERVICE LEVEL MANAGEMENT\n"
  "Fabasoft app.telemetry comes with application management features "
  "for heterogeneous platforms like Microsoft Windows or Red Hat Linux. Threshold "
  "configurations for warnings and critical situations  - using Windows system "
  "counters on Microsoft Windows or SNMP values on Linux-based systems - as well "
  "as predefined and application specific availability checks allow for proactive "
  "handling of issues before they impact production systems.\n\n"

  "SOFTWARE DEVELOPMENT KIT\n"
  "Fabasoft app.telemetry offers an easy to use software development "
  "kit. Software developers can enable their applications for Software Telemetry "
  "with a minimum amount of effort and fully integrate them into Fabasoft app.telemetry. "
  "Contact us for your free software development kit today: "
  "office@fabasoft.com";

// callback function for registering event descriptions
void RepositoryEnumEvents()
{
  APMModuleRegisterEvent(g_module, APM_EVENT_PROCESS_REQUEST, "Process Request", NULL);
  APMModuleRegisterEvent(g_module, APM_EVENT_GET_ARTICLE, "Get Article", NULL);
  APMModuleRegisterEvent(g_module, APM_EVENT_GET_ARTICLE_DATA, "GetArticle::length", "length(bytes)");
}

// register application with APM Software-Telemetry
void RegisterSoftwareTelemetry()
{
  // Register Application
  g_app = APMRegisterApplication("Fabasoft app.telemetry", "Demo", "Backend Services", "Repository", 0);

  // Register Module
  g_module = APMRegisterModule("Repository Handler");
  RepositoryEnumEvents();
}

//------------------------------------------------------------------------------
// ISAPI extension function
//------------------------------------------------------------------------------

BOOL WINAPI GetExtensionVersion(HSE_VERSION_INFO  *pVer)
{
  pVer->dwExtensionVersion = HSE_VERSION;
  strncpy(pVer->lpszExtensionDesc, "Fabasoft app.telemetry Demo IIS Extension", HSE_MAX_EXT_DLL_NAME_LEN);
  return TRUE;
}

BOOL WINAPI TerminateExtension(DWORD dwFlags)
{
  return TRUE;
}

DWORD WINAPI HttpExtensionProc(EXTENSION_CONTROL_BLOCK *pECB)
{
  WebServiceContext context(pECB);
  context.Process();
  return HSE_STATUS_SUCCESS;
}


//------------------------------------------------------------------------------
// DllMain
//------------------------------------------------------------------------------
extern "C"
BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID /*lpReserved*/)
{
  switch (dwReason) {
  case DLL_PROCESS_ATTACH:
    RegisterSoftwareTelemetry();
    _AtlModule.RegisterServer();
    break;
  case DLL_PROCESS_DETACH:
    _AtlModule.UnregisterServer();
    _AtlModule.Term();
    break;
  case DLL_THREAD_ATTACH:
    break;
  case DLL_THREAD_DETACH:
    break;
  }
  return TRUE;
}
