/* ==========================================================================
 * Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
 *
 * All rights reserved. All hardware and software names used are registered
 * trade names and/or registered trademarks of the respective manufacturers.
 *
 * The user of this computer program acknowledges that the above copyright
 * notice, which constitutes the Universal Copyright Convention, will be
 * attached at the position in the function of the computer program which the
 * author has deemed to sufficiently express the reservation of copyright.
 * It is prohibited for customers, users and/or third parties to remove,
 * modify or move this copyright notice.
 * ========================================================================= */

#include <httpext.h>
#include <string>
#include <list>
#include <windows.h>
#include <sstream>
#include <iostream>
#include "base64.h"
#include "softwaretelemetry.h"

#define APM_EVENT_PROCESS_REQUEST  100
#define APM_EVENT_GET_ARTICLE      101
#define APM_EVENT_GET_ARTICLE_DATA 102

extern APMApplicationHandle g_app;
extern APMModuleHandle g_module;
extern const char *g_result_text;

class WebServiceContext
{
private:
  EXTENSION_CONTROL_BLOCK *m_pecb;

public:
  WebServiceContext(EXTENSION_CONTROL_BLOCK *pecb) :
      m_pecb(pecb)
  {
  }

  std::string GetServerVariable(const char *name)
  {
    char buffer[1024];
    DWORD size = sizeof(buffer);
    if (m_pecb->GetServerVariable(m_pecb->ConnID, const_cast<char*>(name), buffer, &size)) {
      return std::string(buffer, size);
    }
    else if (ERROR_INSUFFICIENT_BUFFER==GetLastError()) {
      char *value = (char *)malloc(size);
      if (value) {
        if ((*m_pecb->GetServerVariable)(m_pecb->ConnID, const_cast<char*>(name), value, &size)) {
          std::string result(value, size);
          free(value);
          return result;
        }
        free(value);
      }
    }
    return std::string();
  }

  void Process()
  {
    char bp[1024];
    DWORD reqbytes = sizeof(bp);
    std::string telemetryHeader = GetServerVariable("HTTP_TelemetryContext");
    if (telemetryHeader.length() >= 24) {
      APMContext context;
      if (sizeof(context) == Base64Decoder::Convert(telemetryHeader, context, sizeof(context))) {
        // attach to the Software-Telemetry context of the calling process
        APMAttachContext(g_app, &context);
      }
    }

    // mark start of request processing
    APMEvent(g_module,
      APM_EVENT_PROCESS_REQUEST,
      APM_EVENT_LEVEL_NORMAL,
      APM_FLAG_ENTER);

    std::string result = "nothing to report";

    // read request information
    reqbytes = m_pecb->cbTotalBytes - m_pecb->cbAvailable;
    strncpy(bp, (const char *)m_pecb->lpbData, m_pecb->cbAvailable);
    DWORD cbOffset = m_pecb->cbTotalBytes;
    while (reqbytes > 0) {
      DWORD received = reqbytes;
      if ((*m_pecb->ReadClient)(m_pecb->ConnID, bp + cbOffset, &received)) {
        if (received) {
          reqbytes -= received;
          cbOffset += received;
        }
        else {
          break;
        }
      }
      else {
        break;
      }
    }
    bp[cbOffset] = 0;

    int article=0, offset=0, length=4096;
    if (sscanf(bp, "%d,%d,%d", &article, &offset, &length) >= 1) {
      std::string article_text = GetArticle(article);
      if (offset < article_text.length()) {
        result = article_text.substr(offset, length);
      }
      else {
        result.clear();
      }
    }

    Sleep(2);

    DWORD size = (DWORD)result.length();
    char headerBuffer[512];
    sprintf(headerBuffer, "Content-Type: text/plain\r\nContent-Length: %d\r\n", (int)size);
    APMSyncMark syncMark;
    if (APMGetSyncMark(&syncMark)) {
      std::string sSyncMark = Base64Encoder::Convert(syncMark, sizeof(syncMark));
      strcat(headerBuffer, "TelemetrySyncMark: ");
      strcat(headerBuffer, sSyncMark.c_str());
      strcat(headerBuffer, "\r\n");
    }
    strcat(headerBuffer, "\r\n");
    HSE_SEND_HEADER_EX_INFO HeaderExInfo;
    HeaderExInfo.pszStatus = "200 OK";
    HeaderExInfo.pszHeader = headerBuffer;
    HeaderExInfo.cchStatus = (DWORD)strlen(HeaderExInfo.pszStatus);
    HeaderExInfo.cchHeader = (DWORD)strlen(HeaderExInfo.pszHeader);
    HeaderExInfo.fKeepConn = TRUE;

    // send headers using IIS-provided callback
    //
    // Note - if you want to keep the connection open, set fKeepConn
    // to TRUE *and* provide correct Content-Length: header

    m_pecb->ServerSupportFunction(m_pecb->ConnID, HSE_REQ_SEND_RESPONSE_HEADER_EX, &HeaderExInfo, NULL, NULL);

    (*m_pecb->WriteClient)(m_pecb->ConnID, (void *)result.c_str(), &size, HSE_IO_SYNC);
    APMEvent(g_module, APM_EVENT_PROCESS_REQUEST, APM_EVENT_LEVEL_NORMAL, APM_FLAG_LEAVE);
    APMReleaseContext();
  }


  std::string GetArticle(int article)
  {
    // mark start of work
    APMEvent(g_module,
      APM_EVENT_GET_ARTICLE,
      APM_EVENT_LEVEL_NORMAL,
      APM_FLAG_ENTER);

    // very simple work done in this sample
    std::string result(g_result_text);

    // trace length of result text
    APMParamHandle param = APMParamInit();
    APMParamAdd32(param, (UInt32)result.length());
    APMEventParam(g_module,
      APM_EVENT_GET_ARTICLE_DATA,
      APM_EVENT_LEVEL_DETAIL,
      APM_FLAG_NONE,
      param);
    APMParamFree(param);

    // mark end of work
    APMEvent(g_module,
      APM_EVENT_GET_ARTICLE,
      APM_EVENT_LEVEL_NORMAL,
      APM_FLAG_LEAVE);

    return result;
  }


};
