// ==========================================================================
// $Workfile: MyViewController.m $
// $Revision: 14525 $ of $Date: 2012-11-05 12:49:27 +0100 (Mo, 05 Nov 2012) $
//
// Copyright (c) Fabasoft R&D GmbH, Linz, Austria, 1995-2015.
//
// All rights reserved. All hardware and software names used are registered
// trade names and/or registered trademarks of the respective manufacturers.
//
// The user of this computer program acknowledges that the above copyright
// notice, which constitutes the Universal Copyright Convention, will be
// attached at the position in the function of the computer program which the
// author has deemed to sufficiently express the reservation of copyright.
// It is prohibited for customers, users and/or third parties to remove,
// modify or move this copyright notice.
// ==========================================================================

#import "MyViewController.h"
#import "softwaretelemetry_iOS.h"
#import <QuartzCore/QuartzCore.h>   // required for screenshot

@implementation MyViewController
@synthesize textField;
@synthesize labelResult;
@synthesize switchScreenshot;
@synthesize userName;

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    // Release any cached data, images, etc that aren't in use.
}

#pragma mark - View lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    
	// Do any additional setup after loading the view, typically from a nib.
    
    //NSString *serverUrl = @"http://telemetry.fabasoft.com/web.telemetry";
    NSString *serverUrl = @"https://apmtestclient.fabagl.fabasoft.com/web.telemetry";
    //NSString *serverUrl = @"https://10.10.0.20/web.telemetry";

    APM *apm = [APM initializeWithURL:serverUrl printDebugLogs:true];
    [apm registerApplication:@"Fabasoft app.telemetry" appId:@"iOS Sample" tierName:@"Apple iOS SDK" tierId:@"1"];

    APMModule *apmMod = [apm getModule:@"TestModule"];
    [apmMod registerEvent:900 name:@"Event" parameters:@"text"];
    [apmMod registerEvent:901 name:@"send Message" parameters:@"message"];
    [apmMod registerEvent:910 name:@"Test Event" parameters:@""];
    [apmMod registerEvent:912 name:@"Event with Params" parameters:@"p1;p2;p3"];
    [apmMod registerEvent:920 name:@"Sleep" parameters:@"time_ms"];
    [apmMod registerEvent:930 name:@"Report" parameters:@"reportkey"];
    [apmMod registerEvent:931 name:@"ReportContent" parameters:@"contentsize"];
    [apmMod registerEvent:935 name:@"Render Screenshot" parameters:@""];
    [apmMod registerEvent:940 name:@"Inner Loop" parameters:@"count"];
    
    self.userName = [[NSString alloc] initWithFormat:@"SESSION_%d", arc4random_uniform(10000)];
    [apm registerFilterValue:self.userName value:self.userName];
    [self fireSampleRequests:nil];
    
}

- (void)viewDidUnload
{
    [self setTextField:nil];
    [self setLabelResult:nil];
    [self setSwitchScreenshot:nil];
    [super viewDidUnload];
    // Release any retained subviews of the main view.
    // e.g. self.myOutlet = nil;
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
}

- (void)viewDidAppear:(BOOL)animated
{
    [super viewDidAppear:animated];
}

- (void)viewWillDisappear:(BOOL)animated
{
	[super viewWillDisappear:animated];
}

- (void)viewDidDisappear:(BOOL)animated
{
	[super viewDidDisappear:animated];
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
    // Return YES for supported orientations
    return (interfaceOrientation != UIInterfaceOrientationPortraitUpsideDown);
}

- (IBAction)fireSampleRequests:(id)sender {
    NSLog(@"GUI: fireSampleRequests");
    NSString *nameString = self.textField.text;
    if ([nameString length] == 0) {
        nameString = @"World";
    }
    NSString *greeting = [[NSString alloc] initWithFormat:@"Fired Requests '%@'", nameString];
    self.labelResult.text = greeting;
    
    
    APM *apm = [APM getInstance];
    APMModule *apmMod = [apm getModule:@"TestModule"];
    
    
    // request 1
    NSNumber *ctx = [apm createContextTx:0 filterValue:self.userName];
    [apmMod eventTx:ctx id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"My 1st Event"];
    [apmMod eventTx:ctx id:901 level:APMLogLevelLog flags:APMFlagNone stringValue:greeting];

    [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagEnter|APMFlagWait stringValue:@"50"];
    usleep(50000);
    [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagLeave|APMFlagWait];
    [apm releaseContextTx:ctx];
    
    // request 2
    ctx = [apm createContextTx:0 filterValue:self.userName];
    [apmMod eventTx:ctx id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"1st Event in Request 2"];
    [apmMod eventTx:ctx id:910 level:APMLogLevelNormal flags:APMFlagEnter];
    NSMutableArray *params = [NSMutableArray arrayWithCapacity:3];
    NSNumber *myNumber = [NSNumber numberWithUnsignedInt:222];
    NSNumber *myNumber2 = [NSNumber numberWithUnsignedInt:1234366789];
    //NSNumber *myNumber2 = [NSNumber numberWithFloat:333.321]; // not supported!
    [params addObject:@"1.Param"];
    [params addObject:myNumber];
    [params addObject:myNumber2];
    [apmMod eventTx:ctx id:912 level:APMLogLevelDetail flags:APMFlagNone parameters:params];
    
    [apmMod eventTx:ctx id:912 level:APMLogLevelDetail flags:APMFlagNone parameters:[NSArray arrayWithObjects:@"Test inline Array",myNumber,myNumber2,nil]];
    
    [apmMod eventTx:ctx id:910 level:APMLogLevelNormal flags:APMFlagLeave];
    
    
    [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagEnter|APMFlagWait stringValue:@"200"];
    usleep(200000);
    [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagLeave|APMFlagWait];
    NSData *ctxToken = [apm getContextTx:ctx];
    NSNumber *ctx2 = [apm attachContextTx:0 context:ctxToken];
    [apmMod eventTx:ctx2 id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"Event in Thread 2"];
    NSData *syncToken = [apm getSyncMarkTx:ctx2];
    [apm releaseContextTx:ctx2];

    
    [apm setSyncMarkTx:0 syncMark:syncToken];
    [apmMod eventTx:ctx id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"after Thread 2 done"];
    [apm releaseContextTx:ctx];    
}

- (IBAction)fireManyRequestsClicked:(id)sender {
    uint count = 16;
    uint r = 0;
    long i = 0;
    NSLog(@"GUI: fireManyRequestsClicked %d", count);
    
    //NSString *serverUrl = @"http://apmtestclient.fabagl.fabasoft.com/web.telemetry";
    //APM *apm = [APM initializeWithURL:serverUrl printDebugLogs:true];
    //[apm registerApplication:@"Fabasoft app.telemetry" appId:@"iOS Sample" tierName:@"Apple iOS SDK" tierId:@"1"];
    
    APM *apm = [APM getInstance];
    APMModule *apmMod = [apm getModule:@"TestModule"];
    
    for (r = 1; r <= count; r++) {
        NSNumber *ctx = [apm createContextTx:0 filterValue:self.userName];
        [apmMod eventTx:ctx id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"My 1st Event - fireManyRequestsClicked"];
        [apmMod eventTx:ctx id:900 level:APMLogLevelDebug flags:APMFlagNone stringValue:@"Debug Level Event"];
        [apmMod eventTx:ctx id:901 level:APMLogLevelLog flags:APMFlagNone stringValue:@"Firing many requests with many events"];
        
        // inner loop
        long innerCount = r * 100;
        NSMutableArray *testParams = [NSMutableArray arrayWithCapacity:1];
        [testParams addObject:@(innerCount)];
        [apmMod eventTx:ctx id:940 level:APMLogLevelLog flags:APMFlagEnter parameters:testParams];
        for (i = 0; i < innerCount; i++) {
            [apmMod eventTx:ctx id:910 level:APMLogLevelDetail flags:APMFlagNone];
            NSMutableArray *px = [NSMutableArray arrayWithCapacity:1];
            [px addObject:@(i)];
            [apmMod eventTx:ctx id:940 level:APMLogLevelDetail flags:APMFlagNone parameters:px];
        }
        [apmMod eventTx:ctx id:940 level:APMLogLevelLog flags:APMFlagLeave];
        [apm releaseContextTx:ctx];
    }
    NSString *resultMsg = [[NSString alloc] initWithFormat:@"%d requests fired", count];
    self.labelResult.text = resultMsg;
}

- (IBAction)fireSlowRequestClicked:(id)sender {
    NSLog(@"GUI: Start slow long lasting request");
    APM *apm = [APM getInstance];
    APMModule *apmMod = [apm getModule:@"TestModule"];
    
    NSNumber *ctx = [apm createContextTx:0 filterValue:self.userName];
    self.labelResult.text = @"Slow request started and running ...";
    
    [apmMod eventTx:ctx id:900 level:APMLogLevelLog flags:APMFlagNone stringValue:@"My 1st Event - fireSlowRequestClicked"];
    [apmMod eventTx:ctx id:901 level:APMLogLevelLog flags:APMFlagNone stringValue:@"Firing one long lasting slow request"];
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
        // inner loop
        long innerCount = 100;
        long i;
        NSMutableArray *testParams = [NSMutableArray arrayWithCapacity:1];
        [testParams addObject:@(innerCount)];
        [apmMod eventTx:ctx id:940 level:APMLogLevelLog flags:APMFlagEnter parameters:testParams];
        for (i = 0; i < innerCount; i++) {
            [apmMod eventTx:ctx id:910 level:APMLogLevelDetail flags:APMFlagNone];
            NSMutableArray *px = [NSMutableArray arrayWithCapacity:1];
            [px addObject:@(i)];
            [apmMod eventTx:ctx id:940 level:APMLogLevelDetail flags:APMFlagNone parameters:px];
            [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagEnter|APMFlagWait stringValue:@"1000"];
            usleep(1000000);
            [apmMod eventTx:ctx id:920 level:APMLogLevelDetail flags:APMFlagLeave|APMFlagWait];
        }
        [apmMod eventTx:ctx id:940 level:APMLogLevelLog flags:APMFlagLeave];
        
        [apm releaseContextTx:ctx];
        self.labelResult.text = @"Slow request finished.";
    });
}


- (IBAction)doReport:(id)sender {
    NSLog(@"GUI: Send new Report");
    APM *apm = [APM getInstance];
    APMModule *apmMod = [apm getModule:@"TestModule"];

    NSNumber *ctx = [apm createContextTx:0 filterValue:self.userName];
    
    NSString *repk = [[NSString alloc] initWithFormat:@"REPK_%d", arc4random_uniform(10000)];
    [apmMod eventTx:ctx id:930 level:APMLogLevelLog flags:APMFlagNone stringValue:repk];
    
    NSString *nameString = self.textField.text;
    if ([nameString length] == 0) {
        nameString = [[NSString alloc] initWithFormat:@"Description %@", repk];
    }
    NSString *descr = [[NSString alloc] initWithFormat:@"%@", nameString];
    
    [apm reportTx:ctx filterValue:self.userName reportKey:repk description:descr];
    [apm reportValueTx:ctx reportKey:repk valueKey:@"Rep_Val1" value:@"Report Value 1"];

    [apmMod eventTx:ctx id:900 level:APMLogLevelDetail flags:APMFlagNone stringValue:@"Building Report Content"];
    NSMutableString *strData = [NSMutableString stringWithCapacity:200];
    [strData appendString:@"This is a sample text content sent with the ReportContent function of the Fabasoft app.telemetry SDK for iOS.\r\n"];
    [strData appendString:@"This is the 2nd text line of the sample text content.\r\n"];
    [strData appendString:@"This line contains many special char symbols: ÄÖÜäöü@,;.:-_#'+*°!§$%&/()=?`„¡“¶¢[]|{}≠¿'±‘æ…–∞œ˛÷—Ø°˜\\^£#¬˛÷—ÆŒﬂˆÛ.\r\n"];
    [strData appendString:@"END of sample text content."];
    NSData *data = [strData dataUsingEncoding:NSUTF8StringEncoding];
    NSString *dataSize = [[NSString alloc] initWithFormat:@"%lu", (unsigned long)[data length]];
    APMReportContent *content = [[APMReportContent alloc] initWithData:@"Text-Content.txt" mimeType:@"text/plain" content:data];
    [apmMod eventTx:ctx id:931 level:APMLogLevelDetail flags:APMFlagNone stringValue:dataSize];
    [apm reportContent:repk withContent:content];
    
    if ([self.switchScreenshot isOn]) {
        [apm resumeTransport];
        [apmMod eventTx:ctx id:935 level:APMLogLevelDetail flags:APMFlagEnter|APMFlagWait];
        APMReportContent *screenshot = [APM createScreenshotForView:self.view];
        
        /*if ([[UIScreen mainScreen] respondsToSelector:@selector(scale)]) {
            UIGraphicsBeginImageContextWithOptions(self.view.bounds.size, NO, [UIScreen mainScreen].scale);
        } else {
            UIGraphicsBeginImageContext(self.view.bounds.size);
        }

        [self.view.layer renderInContext:UIGraphicsGetCurrentContext()];
        UIImage *viewImage = UIGraphicsGetImageFromCurrentImageContext();
        UIGraphicsEndImageContext();
        
        // NSData * imageData = UIImagePNGRepresentation(viewImage);
        // APMReportContent *screenshot = [[APMReportContent alloc] initWithData:@"iOS Screenshot" mimeType:@"image/png" content:imageData];

        NSData * imageData = UIImageJPEGRepresentation(viewImage, 0.4);
        APMReportContent *screenshot = [[APMReportContent alloc] initWithData:@"iOS Screenshot" mimeType:@"image/jpeg" content:imageData];
        */
        
        [apmMod eventTx:ctx id:935 level:APMLogLevelDetail flags:APMFlagLeave|APMFlagWait];
        
        NSString *screenshotSize = [[NSString alloc] initWithFormat:@"%lu", (unsigned long)[[screenshot getContent] length]];
        [apmMod eventTx:ctx id:931 level:APMLogLevelDetail flags:APMFlagNone stringValue:screenshotSize];
        [apm reportContent:repk withContent:screenshot];
        self.labelResult.text = [[NSString alloc] initWithFormat:@"Report sent with screenshot: '%@'", descr];
    } else {
        [apm suspendTransport];
        [apmMod eventTx:ctx id:900 level:APMLogLevelDetail flags:APMFlagNone stringValue:@"Skipping Screenshot"];
        self.labelResult.text = [[NSString alloc] initWithFormat:@"Report sent - NO screenshot: '%@'", descr];
    }
    
    [apm releaseContextTx:ctx];
}

- (BOOL)textFieldShouldReturn:(UITextField *)theTextField {
    if (theTextField == self.textField) {
        [theTextField resignFirstResponder];
    }
    return YES;
}
@end
