// ==========================================================================
// $Workfile: APM.h $
// $Revision: 24154 $ of $Date: 2015-10-22 14:38:04 +0200 (Thu, 22 Oct 2015) $
//
// Copyright (c) Fabasoft R&D GmbH, Linz, Austria, 1995-2015.
//
// All rights reserved. All hardware and software names used are registered
// trade names and/or registered trademarks of the respective manufacturers.
//
// The user of this computer program acknowledges that the above copyright
// notice, which constitutes the Universal Copyright Convention, will be
// attached at the position in the function of the computer program which the
// author has deemed to sufficiently express the reservation of copyright.
// It is prohibited for customers, users and/or third parties to remove,
// modify or move this copyright notice.
// ==========================================================================

#import <Foundation/Foundation.h>
#include "APMTransport.h"
#include "APMModule.h"

#import <UIKit/UIKit.h>             // required for screenshot

typedef NS_OPTIONS(UInt8, APMLogLevel) {
    APMLogLevelLog = 0,
    APMLogLevelNormal = 50,
    APMLogLevelDetail = 60,
    APMLogLevelDebug = 70,
};

typedef NS_OPTIONS(UInt8, APMFlag) {
    APMFlagNone = 0,
    APMFlagEnter = 1,
    APMFlagLeave = 2,
    APMFlagWait = 4,
    APMFlagWarning = 8,
    APMFlagError = 16,
    APMFlagProcessInfo = 32,
};

//
//  predefined modules
//
//  applications may use predefined modules and events
//  to use a predefined module, register a APMModule with
//  the respective name. Event descriptions need not be registered.
//

//NSString *MODULE_NAME_OS = "Operating System";
typedef NS_OPTIONS(UInt64, APMEventModuleOS) {
    EVENT_OS_CREATE = 0x0000,
    EVENT_OS_OPEN = 0x0001,
    EVENT_OS_DELETE = 0x0002,
    EVENT_OS_READ = 0x0003,
    EVENT_OS_WRITE = 0x0004,
    EVENT_OS_CLOSE = 0x0005,
    EVENT_OS_FLUSH = 0x0006,
    EVENT_OS_RENAME = 0x0007,
    EVENT_OS_COPY = 0x0008,
    EVENT_OS_MOVE = 0x0009,
    EVENT_OS_LOCK = 0x000A,
    EVENT_OS_UNLOCK = 0x000B,
    EVENT_OS_EXECUTE = 0x0100,
    EVENT_OS_WAIT = 0x0200,
    EVENT_OS_SLEEP = 0x0201,

    EVENT_OS_CRITICALSECTION_WAIT = 0x0300,
    EVENT_OS_CRITICALSECTION_ENTER = 0x0301,
    EVENT_OS_CRITICALSECTION_LEAVE = 0x0302,
};

// NSString *MODULE_NAME_XMLHTTP = "XMLHttpRequest";
typedef NS_OPTIONS(UInt64, APMEventModuleXMLHTTPRequest) {
    EVENT_XMLHTTP_SEND = 0,
    EVENT_XMLHTTP_LOADING = 1,
    EVENT_XMLHTTP_LOADED = 2,
    EVENT_XMLHTTP_INTERACTIVE = 3,
    EVENT_XMLHTTP_COMPLETE = 4,
    EVENT_XMLHTTP_PROCESSING = 10,
    EVENT_XMLHTTP_PROCESSED = 11,
};

//
// predefined events:
//
// the following events are implicitly available under
// all modules
//
typedef NS_OPTIONS(UInt64, APMEventModuleDefault) {
    EVENT_ERROR = 0x10001000,
    EVENT_WARNING = 0x10001001,
    EVENT_INFO = 0x10001002,
    EVENT_TRACE = 0x10001003,
};

@interface APM : NSObject {
@private
    APMTransport *m_transport;
    NSMutableDictionary *m_modules;
}

+ (APM*)initializeWithTransport:(APMTransport*)transport;

+ (APM*)initializeWithURL:(NSString *)url;
+ (APM*)initializeWithURL:(NSString *)url printDebugLogs:(Boolean)debugLog;

+ (APM*)getInstance;

+ (APMModule*)getModule:(NSString *)name;

+ (APMReportContent*)createScreenshotForView:(UIView *)view;

+ (NSString*)encodeBase64:(NSData*)data;

+ (NSData*)decodeBase64:(NSString*)string;

+ (NSData*)generateToken;

+ (NSData*)generateTokenWithLevel:(UInt8)level;

- (id)initWithTransport:(APMTransport*)transport;

- (id)setTransport:(APMTransport*)transport;

- (APMTransport*)getTransport;

- (id)setTransportUrl:(NSString*)url;

- (Boolean)isTransportSuspended;

- (void)suspendTransport;

- (void)resumeTransport;

- (void)setInsecureCertificatesAllowed:(BOOL)allowed;

- (void)flushTransport;

- (Boolean)isConnected;

- (UInt32)registerApplication:(NSString *)applicationName
                        appId:(NSString *)applicationId
                     tierName:(NSString *)applicationTierName
                       tierId:(NSString *)applicationTierId;

- (void)registerFilterValue:(NSString *)name
                      value:(NSString *)value;

- (APMModule*)getModule:(NSString*)name;

- (NSNumber*)createContextTx:(NSNumber*)thread
                 filterValue:(NSString *)filter;

- (NSNumber*)attachContextTx:(NSNumber*)thread
                     context:(NSData*)ctx;

- (NSData*)getContextTx:(NSNumber*)thread;

- (void)syncContextTx:(NSNumber*)thread
              context:(NSData*)ctx;

- (void)releaseContextTx:(NSNumber*)thread;

- (NSData*)getSyncMarkTx:(NSNumber*)thread;

- (void)setSyncMarkTx:(NSNumber*)thread
             syncMark:(NSData*)syncmark;

- (void)reportTx:(NSNumber*)thread
     filterValue:(NSString*)filterValue
       reportKey:(NSString*)reportKey
     description:(NSString*)description;

- (void)reportValueTx:(NSNumber*)thread
            reportKey:(NSString*)reportKey
             valueKey:(NSString*)key
                value:(NSString*)value;

- (void)reportContent:(NSString*)reportKey
          withContent:(APMReportContent*)content;

- (Boolean)hasActiveContext:(NSNumber*)thread
                      level:(UInt8)level;

@end

