// ==========================================================================
// $Workfile: APMTransport.h $
// $Revision: 24154 $ of $Date: 2015-10-22 14:38:04 +0200 (Thu, 22 Oct 2015) $
//
// Copyright (c) Fabasoft R&D GmbH, Linz, Austria, 1995-2015.
//
// All rights reserved. All hardware and software names used are registered
// trade names and/or registered trademarks of the respective manufacturers.
//
// The user of this computer program acknowledges that the above copyright
// notice, which constitutes the Universal Copyright Convention, will be
// attached at the position in the function of the computer program which the
// author has deemed to sufficiently express the reservation of copyright.
// It is prohibited for customers, users and/or third parties to remove,
// modify or move this copyright notice.
// ==========================================================================

#import <Foundation/Foundation.h>
#import "APMReportContent.h"

//
// app.telemetry transport entry ids
//
typedef NS_OPTIONS(UInt16, APMEntryIds) {
    SWT_ENTRY_GET_CONTEXT = 5,
    SWT_ENTRY_ATTACH_CONTEXT = 6,
    SWT_ENTRY_EVENT = 7,
    SWT_ENTRY_CREATE_CONTEXT = 8,
    SWT_ENTRY_RELEASE_CONTEXT = 9,
    SWT_ENTRY_GET_SYNC_MARK = 15,
    SWT_ENTRY_SET_SYNC_MARK = 16,
    SWT_ENTRY_SYNC_CONTEXT = 17,
    SWT_ENTRY_REPORT = 18,
    SWT_ENTRY_REPORT_VALUE = 19,
};

static UInt32 APM_MAX_EVENT_LIMIT = 1000;
static UInt32 APM_MAX_FLUSH_INTERVAL = 20; // sec

@interface APMTransport : NSObject

- (id)initWithURL:(NSString*)url;
- (id)initWithURL:(NSString*)url printDebugLogs:(Boolean)debugLog;

- (void)cleanupTransport;

- (void)setUrl:(NSString*)url;

- (void)APMRegisterApplication:(NSString *)applicationName
                         appId:(NSString *)applicationId
                      tierName:(NSString *)applicationTierName
                        tierId:(NSString *)applicationTierId;

- (UInt32)APMRegisterModule:(NSString *)name;

- (void)APMRegisterEvent:(UInt32)module
                   event:(UInt64)eventid
                    name:(NSString *)name
                   param:(NSString *)params;

- (void)APMRegisterFilterValue:(NSString *)name
                         value:(NSString *)value;

- (void)APMReportContent:(NSString*)reportKey
                 content:(APMReportContent*)content;

- (UInt8)getSessions:(NSString *)filter
            sessions:(NSMutableArray *)sessions;

- (UInt8)getApplicationEventLevel;

- (void)flush;

- (Boolean)isConnected;

- (NSNumber*)getNextThreadId;

- (NSNumber*)getThreadLevel:(NSNumber*)thread;

- (void)putThreadLevel:(NSNumber*)thread
                 level:(UInt8)level;

- (void)removeThreadLevel:(NSNumber*)thread;

- (NSMutableArray*)createItem:(UInt16)type
                       thread:(NSNumber*)thread;

- (void)addItem:(NSArray *)item;

- (Boolean)isInternetConnected;

- (void)doSuspend;

- (void)doResume;

- (void)setInsecureCertificatesAllowed:(BOOL)allowed;

- (Boolean)isSuspended;

@property (nonatomic, strong) NSString* defaultUserAgent;

@end
