/**
@license Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1988-2015.

All rights reserved. All hardware and software names used are registered
trade names and/or registered trademarks of the respective manufacturers.

The user of this computer program acknowledges that the above copyright
notice, which constitutes the Universal Copyright Convention, will be
attached at the position in the function of the computer program which the
author has deemed to sufficiently express the reservation of copyright.
It is prohibited for customers, users and/or third parties to remove,
modify or move this copyright notice.
*/
/*jslint browser: true, sub: true, undef: true, maxerr: 5000,
 eqeqeq: true, regexp: true, newcap: true, immed: true */
/*global window: true, $: true */

(function($) {
  var ts, slice = [].slice, re_arg = new RegExp('\\{\\d+\\}', 'gi');
  ts = Date.now ? Date.now : function () {
    return (new Date()).getTime();
  };
  function arg(str /*, args*/) {
    var args = slice.call(arguments, 1);
    return str.replace(re_arg, function(match) {
      return args[parseInt(match.substr(1, match.length - 2), 10)];
    });
  }

  $.widget("apm.speedtest", {
    options: {
      lang: {
        header: 'Fabasoft app.telemetry 2015 Speed Test',
        footer: 'Copyright \u00A9 Fabasoft R&D GmbH, Linz, Austria, 1988-2015.',
        headings: [ 'Test', 'Average', 'Worst', 'Best', 'Size' ],
        rows: [ 'Round-trip time', 'Text download', 'Script download', 'Data download', 'Text upload' ],
        failed: 'failed',
        buttons: {
          start: 'Start speed test',
          abort: 'Abort speed test'
        },
        phase: {
          init: 'Initializing test ... ',
          init_failed: 'failed (service not available)!',
          roundtrip: 'round-trip time',
          upload: 'upload bandwidth',
          upload_speed: 'upload speed',
          download_speed: 'download speed',
          download_speed_data: 'download speed (data)',
          download_speed_script: 'download speed (script)',
          download_speed_text: 'download speed (text)',
          download: 'download bandwidth',
          done: 'done'
        },
        unit: {
          bps: 'bit/s',
          kbps: 'Kbit/s',
          mbps: 'Mbit/s',
          gbps: 'Gbit/s',
          bytes: 'B',
          kbytes: 'KB',
          mbytes: 'MB',
          ms: 'ms',
          requests: 'requests'
        },
        progress_x_of_y_size: '{0} / {1} (size: {2})',
        progress_x_of_y: '{0} / {1}',
        test_running: 'Test Running...',
        testing_what_progress: 'Testing {0} ... {1}'
      },
      web_telemetry: "web.telemetry",
      // --------- CONFIGURATION OF LIMITS --------------------
      timeout: 10000,
      // (WARNING: Browser may crash on too large text/script downloads of e.g. 10MB)
      upload_limit: 500,   // KB
      download_limits: {   // KB
        text:   500,
        script: 1000,
        data:   2000
      },
      maxIterations: 5,    // increasing the default value of 5 will increase test run duration
      maxPingCount: 40
      /* TEST EXECUTION ORDER and PROGRESS distribution
       1. PING Test:                      10%
       2. DOWNLOAD:
         2.2: download text:              20%
         2.3: download script:            20%
         2.4: download data/img:          20%
       3. UPLOAD:
         3.1: check upload bandwidth:      5%
         3.2: upload                      20%
      */
    },
    curxhr: {
      xhr: null,
      $if: null,
      $scr: null,
      $img: null
    },
    rowMap: ['rtt', 'txtdl', 'scrdl', 'datadl', 'txtup'],
    colMap: ['avg', 'worst', 'best', 'size'],
    results: {
      'rtt': {
        'avg': null,
        'worst': null,
        'best': null,
        'size': null
      },
      'txtdl': {
        'avg': null,
        'worst': null,
        'best': null,
        'size': null
      },
      'scrdl': {
        'avg': null,
        'worst': null,
        'best': null,
        'size': null
      },
      'datadl': {
        'avg': null,
        'worst': null,
        'best': null,
        'size': null
      },
      'txtup': {
        'avg': null,
        'worst': null,
        'best': null,
        'size': null
      }
    },
    cancelTest: false,
    initialized: false,
    _create: function() {
      var _ = this, o = _.options, el = _.element;
      el.addClass('apm-speedtest');
      el.find('header').text(o.lang.header);
      el.find('footer').text(o.lang.footer);
      el.find('table')
        .find('thead th')
          .each(function(idx, th) {
            $(th).text(o.lang.headings[idx]);
          })
        .end()
        .find('tbody tr')
          .each(function(idx, tr) {
            $(tr).find('td').each(function(tdidx, td) {
              if (tdidx > 0) {
                _.results[_.rowMap[idx]][_.colMap[tdidx - 1]] = $(td);
              }
              else {
                $(td).text(o.lang.rows[idx]);
              }
            });
          });
      el.find('button')
        .each(function() {
          var $t = $(this), action = $t.attr('data-action');
          switch(action) {
          case 'start':
            _.$start = $t;
            _.$start.text(o.lang.buttons.start);
            break;
          case 'abort':
            _.$abort = $t;
            _.$abort.text(o.lang.buttons.abort);
            break;
          }
        });
      el.find('.apm-speedtest-running').text(o.lang.test_running);
      _.$progressBar = el.find('.apm-speedtest-progressbar div');
      _.$progressPane = el.find('.apm-speedtest-progress-pane');
      _.$progressMessage = el.find('.apm-speedtest-progress-message');
      el.find('.apm-speedtest-button-pane').css('visibility', 'visible');

      var web_telemetry = o.web_telemetry, timeout = o.timeout, upload_limit = o.upload_limit,
        download_limits = o.download_limits, maxIterations = o.maxIterations, maxPingCount = o.maxPingCount,
        curxhr = _.curxhr;

      var start,
        tx = ts(), latency = 0;

      function formatSpeed(speed) {
        if (speed > 0 && speed < 1000000000000) {
          if (speed < 1000) {
            speed = Math.round(speed) + " " + o.lang.unit.bps;
          }
          else if (speed < 1000000) {
            speed = Math.round(speed / 1000) + " " + o.lang.unit.kbps;
          }
          else if (speed < 1000000000) {
            speed = (speed / 1000000).toFixed(1) + " " + o.lang.unit.mbps;
          }
          else {
            speed = (speed / 1000000000).toFixed(1) + " " + o.lang.unit.gbps;
          }
          return speed;
        }
        return o.lang.failed;
      }
      function formatSize(size) {
        var exp, s;
        exp = size.toFixed(0).length;
        if (exp <= 3) {
          s = size.toFixed(0) + " " + o.lang.unit.bytes;
        }
        else if (exp <= 6) {
          s = (size / 1000).toFixed(0) + " " + o.lang.unit.kbytes;
        }
        else {
          s = (size / 1000000).toFixed(0) + " " + o.lang.unit.mbytes;
        }
        return s;
      }
      function formatSizeCount(iterations, size) {
        return iterations + " x " + formatSize(size);
      }
      function setProgress(value, x, y, size) {
        _.$progressBar.width(Math.ceil(value) + "%");
        if (arguments.length > 2) {
          var text = arguments.length > 3 ? o.lang.progress_x_of_y_size : o.lang.progress_x_of_y;
          _._setProgressMessage(arg(text, x, y, size));
        }
        else {
          _._setProgressMessage('');
        }
      }
      function enableStartBtn() {
        _.$start.button('enable');
      }

      function testsDone() {
        _._abort();
        _._setPhase(o.lang.phase.done);
        setProgress(100);
        _.$progressPane.fadeOut('slow', function() {
          _.$progressPane.css({ visibility: 'hidden', display: 'block' });
        });
        _.$abort.button('disable');
        window.setTimeout(enableStartBtn, 1000);
      }

      // Test #3: uploadTest (calling testsDone)
      function uploadTest() {
        var content = $(document.documentElement).html().substr(0, 1000), start, startPhase = ts(), end, iterations = 0, speeds = [];

        content = content.concat(content, content, content, content, content, content, content, content, content); //10K length
        _._setPhase(o.lang.phase.upload);
        function err(xhr) {
          if (!_.cancelTest) {
            _._setFailed(_.results.txtup);
            content = "";
            window.setTimeout(testsDone, 10);
          }
        }
        function finishUpload(speeds) {
          var i = speeds.length, min, max, avg = 0, size = content.length;
          if (!_.cancelTest) {
            content = "";
            min = max = speeds[0];
            while (i--) {
              min = Math.min(min, speeds[i]);
              max = Math.max(max, speeds[i]);
              avg += speeds[i];
            }
            avg /= speeds.length;
            _.results.txtup.avg.text(formatSpeed(avg * 8));
            _.results.txtup.worst.text(formatSpeed(min * 8));
            _.results.txtup.best.text(formatSpeed(max * 8));
            _.results.txtup.size.text(formatSizeCount(iterations, size));
            window.setTimeout(testsDone, 10);
          }
        }
        function request(handler) {
          start = ts();
          curxhr.xhr = $.ajax({
            async: true,
            global: false,
            cache: false,
            dataType: "text",
            timeout: 10000,
            type: "POST",
            url: web_telemetry + "?SpeedUpload&tx=" + tx++,
            processData: false,
            data: content,
            success: handler,
            error: err
          });
        }
        function succ() {
          if (!_.cancelTest) {
            end = ts();
            var diff = (end - start - latency) / 1000, speed = content.length / diff;
            speeds.push(speed);
            iterations++;
            if (iterations < maxIterations && ((end - startPhase) < timeout)) {
              setProgress(80 + 4 * iterations, iterations, maxIterations, formatSize(content.length));
              request(succ);
            }
            else {
              setProgress(99);
              finishUpload(speeds);
            }
          }
        }
        function ramp() {
          end = ts();
          var diff = (end - start - latency) / 1000, speed = content.length / diff, limit = upload_limit * 1000;
          iterations++;
          if (!_.cancelTest && diff < 1 && iterations < 6 && content.length < limit && ((end - startPhase) < timeout)) {
            if (diff < 0.4) {
              content = content.concat(content, content, content, content, content, content, content, content, content); //x10 length
            }
            else if (diff < 0.7) {
              content = content.concat(content, content, content, content); //x5 length
            }
            else {
              content = content.concat(content); //x2 length
            }
            if (content.length >= limit) {
              content = content.substr(0, limit);
              iterations = 6; // stop bandwith checks when limit is reached
            }
            setProgress(75 + 0.8 * iterations, iterations, 6, formatSize(content.length));
            request(ramp);
          }
          else {
            if (!_.cancelTest) {
              _._setPhase(o.lang.phase.upload_speed);
              iterations = 0;
              setProgress(80, 0, 6, formatSize(content.length));
              request(succ);
            }
          }
        }

        // uploadTest main code
        if (!_.cancelTest) {
          request(ramp);
        }
      } // end uploadTest

      // Test #2: downloadTest (ramp, succ_text, succ_script, succ_img) - next: uploadTest
      function downloadTest() {
        var start, startPhase, end, iterations = 0, maxSize = 10, speeds = [], curSize;    // starting size 10KB
        _._setPhase(o.lang.phase.download_speed);

        function finishDownload(speeds, pc, elem, cb) {
          var i = speeds.length, min, max, avg = 0;
          if (!_.cancelTest) {
            min = max = speeds[0];
            while (i--) {
              min = Math.min(min, speeds[i]);
              max = Math.max(max, speeds[i]);
              avg += speeds[i];
            }
            avg /= speeds.length;
            elem.avg.text(formatSpeed(avg * 8));
            elem.worst.text(formatSpeed(min * 8));
            elem.best.text(formatSpeed(max * 8));
            elem.size.text(formatSizeCount(iterations, curSize * 1024));
            _.$progressBar.width(pc + '%');
          }
        }
        function succ_img() { // download test data
          if (!_.cancelTest) {
            end = ts();
            var diff = (end - start - latency) / 1000, speed = (curSize * 1024) / diff;
            speeds.push(speed);
            iterations++;
            if (iterations < maxIterations && ((end - startPhase) < timeout)) {
              setProgress(55 + iterations * 4, iterations, maxIterations, formatSize(curSize * 1024));
              start = ts();
              curxhr.$img.attr('src', web_telemetry + "?SpeedDownload&size=" + curSize + "&type=bmp&tx=" + tx++);
            }
            else {
              finishDownload(speeds, 75, _.results.datadl);
              if (!_.cancelTest) {
                window.setTimeout(uploadTest, 10);
              }
              setProgress(75);
              curxhr.$img.unbind('load').attr('src', 'about:blank').remove();
              delete curxhr.$img;
            }
          }
        }
        function create_script_elem(handler) {
          if (!_.cancelTest) {
            curxhr.$scr = document.createElement('script');
            curxhr.$scr.type = 'text/javascript';
            curxhr.$scr = $(curxhr.$scr);
            curxhr.$scr[0].onload = curxhr.$scr[0].onreadystatechange = function() {
              if ( !this.done && (!this.readyState || this.readyState === "loaded" || this.readyState === "complete") ) {
                this.done = true;
                handler();
                // Handle memory leak in IE
                var scr = this;
                window.setTimeout(function() {
                  scr.onload = scr.onreadystatechange = null;
                  if (scr.parentNode) {
                    scr.parentNode.removeChild(scr);
                  }
                }, 20);
              }
            };
            curxhr.$scr[0].done = false;
            start = ts();
            curxhr.$scr.attr('src', web_telemetry + "?SpeedDownload&size=" + curSize + "&type=script&tx=" + tx++);
          document.body.appendChild(curxhr.$scr[0]);
          }
        }
        function succ_script() {
          if (!_.cancelTest) {
            end = ts();
            var diff = (end - start - latency) / 1000, speed = (curSize * 1024) / diff;
            speeds.push(speed);
            iterations++;
            if (iterations < maxIterations && ((end - startPhase) < timeout)) {
              setProgress(35 + iterations * 4, iterations, maxIterations, formatSize(curSize * 1024));
              create_script_elem(succ_script);
            }
            else {
              finishDownload(speeds, 55, _.results.scrdl);
              speeds = [];
              iterations = 0;
              curxhr.$scr.remove();
              delete curxhr.$scr;
              curxhr.$img = $('<img />');
              curxhr.$img
                .css({
                  position: 'absolute',
                  left: -10000,
                  top: -10000
                })
                .appendTo(document.body)
                .bind('load', succ_img);

              //prepare 1st request for succ_img
              _._setPhase(o.lang.phase.download_speed_data);
              curSize = (maxSize > download_limits.data) ? download_limits.data : maxSize;
              setProgress(55, 0, 5, formatSize(curSize * 1024));
              start = ts();
              startPhase = start;
              curxhr.$img.attr('src', web_telemetry + "?SpeedDownload&size=" + curSize + "&type=bmp&tx=" + tx++);
            }
          }
        }
        function succ_text() {
          if (!_.cancelTest) {
            end = ts();
            var diff = (end - start - latency) / 1000, speed = (curSize * 1024) / diff;
            speeds.push(speed);
            if (iterations < maxIterations && ((end - startPhase) < timeout)) {
              setProgress(15 + iterations * 4, iterations, maxIterations, formatSize(curSize * 1024));
              iterations++;
              window.setTimeout(function() {
                start = ts();
                if (curxhr && curxhr.$if) {
                  curxhr.$if.attr('src', web_telemetry + "?SpeedDownload&size=" + curSize + "&type=text&tx=" + tx++);
                }
              }, 50);
            }
            else {
              finishDownload(speeds, 35, _.results.txtdl);
              speeds = [];
              iterations = 0;
              curxhr.$if.unbind('load').attr('src', 'about:blank').remove();
              delete curxhr.$if;
              //prepare 1st request for succ_script
              _._setPhase(o.lang.phase.download_speed_script);
              curSize = (maxSize > download_limits.script) ? download_limits.script : maxSize;
              setProgress(35, 0, maxIterations, formatSize(curSize * 1024));
              startPhase = ts();
              create_script_elem(succ_script);
            }
          }
        }
        function ramp() { // test download bandwidth
          end = ts();
          var diff = (end - start - latency) / 1000, speed = (maxSize * 1024) / diff;
          if (!_.cancelTest && diff < 1 && iterations < 7 && (maxSize < download_limits.data) && ((end - startPhase) < timeout)) {
            if (diff < 0.4) {
              maxSize *= 10;
            }
            else if (diff < 0.7) {
              maxSize *= 5;
            }
            else {
              maxSize *= 2;
            }
            if (maxSize >= download_limits.data) {
              curSize = maxSize = download_limits.data;
              iterations = 7; // stop bandwith checks when limit is reached
              setProgress(15, 7, 7, formatSize(curSize * 1024));
            } else {
              curSize = maxSize;
              setProgress(10 + (iterations + 1) * 0.7, (iterations + 1), 7, formatSize(curSize * 1024));
            }

          } else {
            iterations = 0;
            curxhr.$if.unbind('load');

            if (!_.cancelTest) {
              //prepare 1st request for succ_text - using text speed max half bandwidth
              if (maxSize > 2*download_limits.text) {
                curSize =  download_limits.text;
              } else {
                curSize = maxSize / 2;
              }
              curxhr.$if.bind('load', succ_text);
              _._setPhase(o.lang.phase.download_speed_text);
              setProgress(15, 0, maxIterations, formatSize(curSize * 1024));
              startPhase = ts();
            }
          }

          if (!_.cancelTest) {
            iterations++;
            start = ts();
            curxhr.$if.attr('src', web_telemetry + "?SpeedDownload&size=" + curSize + "&type=text&tx=" + tx++);
          }
        }

        // downloadTest main code
        if (!_.cancelTest) {
          _._setPhase(o.lang.phase.download);
          curxhr.$if = $('<iframe></iframe>');
          curxhr.$if
            .css({
              position: 'absolute',
              left: -10000,
              top: -10000
            })
            .appendTo(document.body)
            .bind('load', ramp);
          setProgress(10, 0, 7, formatSize(maxSize * 1024));
          start = ts();
          startPhase = start;
          curxhr.$if.attr('src', web_telemetry + "?SpeedDownload&size=" + maxSize + "&type=text&tx=" + tx++);
        }
      } // end downloadTest

      // Test #1: pingTest - next is downloadTest
      function pingTest() {
        var iterations = 0, start = ts(), startPhase = start, durations = [];
        _._setPhase(o.lang.phase.roundtrip);

        function formatLatency(latency) {
          return latency.toFixed(1) + " " + o.lang.unit.ms;
        }
        function finishLatency(durations) {
          var i = durations.length, min, max, avg = 0;
          min = max = durations[0];
          while (i--) {
            min = Math.min(min, durations[i]);
            max = Math.max(max, durations[i]);
            avg += durations[i];
          }
          avg /= durations.length;
          _.results.rtt.avg.text(formatLatency(avg));
          _.results.rtt.worst.text(formatLatency(max));
          _.results.rtt.best.text(formatLatency(min));
          _.results.rtt.size.text(durations.length + " " + o.lang.unit.requests);
          setProgress(10);
          curxhr.xhr = null;
          if (!_.cancelTest) {
            window.setTimeout(downloadTest, 10);
          }
        }
        function err() {
          curxhr.xhr = null;
          if (!_.cancelTest) {
            _._setFailed(_.results.rtt);
            setProgress(10);
            window.setTimeout(downloadTest, 10);
          }
        }
        function request(handler) {
          start = ts();
          curxhr.xhr = $.ajax({
            async: true,
            global: false,
            cache: false,
            dataType: "text",
            timeout: 10000,
            type: "POST",
            url: web_telemetry + "?SpeedUpload&tx=" + tx++,
            success: handler,
            error: err
          });
        }
        function succ() {
          var avg = 0, i, now, last;
          if (!_.cancelTest) {
            now = ts();
            ++iterations;
            durations.push(now - start);
            if (iterations < maxPingCount && ((now - startPhase) < (timeout / 2))) {
              setProgress(iterations * 0.2, iterations, maxPingCount);
              request(succ);
            }
            else if (iterations > 3){ //calculate latency
              finishLatency(durations);
            }
            else {
              err();
            }
          }
        }

        // pingTest main code:
        request(succ);
      } // end pingTest

      function checkWebAPIsuccess() {
        _.$progressMessage.text(o.lang.phase.init + o.lang.phase.done);
        _.$progressPane.css('visibility', 'hidden');
        _.initialized = true;
        _.$start.button('enable');
      }
      function checkWebAPIfail() {
        _.$progressMessage.text(o.lang.phase.init + o.lang.phase.init_failed);
      }
      function checkWebAPI() {
        $.ajax({
          async: true,
          global: false,
          cache: false,
          timeout: 5000,
          type: "GET",
          url: web_telemetry,
          success: checkWebAPIsuccess,
          error: checkWebAPIfail
        });
      }

      // start button: run pingTest (which will call other tests)
      _.$start.button().bind('click', function() {
        if ($(this).hasClass("ui-state-disabled") || _.initialized === false) {
          return;
        }
        _.$progressPane.css('visibility', 'visible');
        _.$start.button('disable');
        _.$abort.button('enable');
        setProgress(0, 0, maxPingCount);
        el.find('.apm-speedtest-running').show();
        _.$progressBar.parent().show();
        _._clearFields(_.results.rtt, '');
        _._clearFields(_.results.txtdl, '');
        _._clearFields(_.results.scrdl, '');
        _._clearFields(_.results.datadl, '');
        _._clearFields(_.results.txtup, '');
        tx = ts();
        latency = 0;
        _.cancelTest = false;
        pingTest();
      }).button('disable');

      // abort all tests
      _.$abort.button().bind('click', function() {
        testsDone();
      }).button('disable');

     // test availability of webAPI
      _.$progressMessage.text(o.lang.phase.init);
      el.find('.apm-speedtest-running').hide();
      _.$progressBar.parent().hide();
      _.$progressPane.css('visibility', 'visible');
      checkWebAPI();
    },
    _destroy: function() {
      this.element.removeClass('apm-speedtest');
      $.each(this.results, function(idx, row) {
        $.each(row, function(val) {
          row[val] = null;
        });
      });
      this.$start.button('destroy').unbind('click');
      delete this.$start;
      this.$abort.button('destroy').unbind('click');
      delete this.$abort;
      delete this.$progressBar;
      delete this.$progressMessage;
      delete this.$progressPane;
    },
    _abort: function() {
      var curxhr = this.curxhr;
      this.cancelTest = true;
      if (curxhr.xhr) {
        try {
          curxhr.xhr.abort();
        }
        catch(_) { /* */ }
        curxhr.xhr = null;
      }
      if (curxhr.$if) {
        curxhr.$if.remove();
        curxhr.$if = null;
      }
      if (curxhr.$scr) {
        curxhr.$scr.remove();
        curxhr.$scr = null;
      }
      if (curxhr.$img) {
        curxhr.$img.remove();
        curxhr.$img = null;
      }
    },
    _setFailed: function(elem) {
      this._clearFields(this.options.lang.failed);
    },
    _clearFields: function(elem, val) {
      elem.avg.text(val);
      elem.worst.text(val);
      elem.best.text(val);
      elem.size.text(val);
    },
    _setPhase: function(phase) {
      this.phase = phase;
      this._updateProgressMessage();
    },
    _setProgressMessage: function(progress) {
      this.progress = progress;
      this._updateProgressMessage();
    },
    _updateProgressMessage: function() {
      this.$progressMessage.text(arg(this.options.lang.testing_what_progress, this.phase, this.progress));
    }
  });
}(jQuery));
