#!/bin/bash
# ===========================================================================
# Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
#
# All rights reserved. All hardware and software names used are registered
# trade names and/or registered trademarks of the respective manufacturers.
#
# The user of this computer program acknowledges that the above copyright
# notice, which constitutes the Universal Copyright Convention, will be
# attached at the position in the function of the computer program which the
# author has deemed to sufficiently express the reservation of copyright.
# It is prohibited for customers, users and/or third parties to remove,
# modify or move this copyright notice.
# ==========================================================================

if [ "$1" != "-s" ]; then
  echo
  echo "  Fabasoft app.telemetry Server Setup"
  echo
  echo "  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1988-2015."
  echo
fi

# --------------------------------------------------------------------------
#  Initialize
# --------------------------------------------------------------------------
if [ $UID != 0 ]; then
  echo "This script must be run by 'root'."
  exit 1
fi

ARCH=`uname -i`
if [ "$ARCH" != "x86_64" ]; then
  echo "Unsupported architecture $ARCH"
  exit 1
fi

if [ -f /etc/os-release ] ; then
  . /etc/os-release
  if [ "$ID" = "centos" -o "$ID" = "rhel" ] ; then
    V=$(echo "$VERSION_ID" | cut -d "." -f -1)
    if [ "$V" = "7" ] ; then
      RPMPATH="LINUX_EL7_X64"
      OS="EL7"
      DEF="A"
    else
      echo "Unsupported operating system in /etc/issue."
      exit 1
    fi
  else
    echo "Unsupported operating system in /etc/issue."
    exit 1
  fi
elif grep -i "Red Hat.*release 5\.[0-9]\|RHEL.*release 5\.[0-9]\|CentOS.*release 5\.[0-9]" /etc/issue > /dev/null ; then
  RPMPATH="LINUX_EL5_X64"
  OS="EL5"
  DEF="B"
elif grep -i "Red Hat.*release 6\.[0-9]\|RHEL.*release 6\.[0-9]\|CentOS.*release 6\.[0-9]" /etc/issue > /dev/null ; then
  RPMPATH="LINUX_EL6_X64"
  OS="EL6"
  DEF="A"
else
  echo "Unsupported operating system in /etc/issue."
  exit 1
fi

# --------------------------------------------------------------------------
#  Installed Packages?
# --------------------------------------------------------------------------
INST_TYPE="INSTALL"
SERVERSETUP=no

AGENT_INST=0
rpm -q apptelemetryagent >/dev/null 2>&1
if [[ "$?" -eq 0 ]]; then
  AGENT_INST=1
  INST_TYPE="UPGRADE"
fi

WEBAPI_INST=0
rpm -q apptelemetryweb >/dev/null 2>&1
if [ "$?" -eq 0 ]; then
  WEBAPI_INST=1
  INST_TYPE="UPGRADE"
fi

SERVER_INST=0
rpm -q apptelemetryserver >/dev/null 2>&1
if [ "$?" -eq 0 ]; then
  SERVER_INST=1
  INST_TYPE="UPGRADE"
fi

# --------------------------------------------------------------------------
#  Check Code Signing Certificate
# --------------------------------------------------------------------------
if [ ! $(rpm -qa "gpg-pubkey*" | grep gpg-pubkey-00744229-5629ed42) ] ; then
  echo
  echo "To verify the authenticity of the Fabasoft app.telemetry rpm packages"
  echo "download the public PGP key from"
  echo "  https://download.fabasoft.com/keys/RPM-PGP-KEY-Fabasoft-RD-2015"
  echo "and install the key using"
  echo "  rpmkeys --install RPM-PGP-KEY-Fabasoft-RD-2015"
fi

# --------------------------------------------------------------------------
#  Upgrade
# --------------------------------------------------------------------------
if [ "$INST_TYPE" = "UPGRADE" ] ; then
  PACKAGES=""
  while true; do
    echo
    echo "The following Fabasoft app.telemetry packages are already installed:"
    if [ $SERVER_INST -eq 1 ]; then
      if [ "$OS" = "EL5" ]; then
        echo "  - WARNING: Fabasoft app.telemetry Server is not supported any longer for this platform and will not be updated! (you should update to RHEL/CentOS 7.x)"
      else
        echo "  - Fabasoft app.telemetry Server"
        PACKAGES="$PACKAGES Server/$RPMPATH/apptelemetryserver*"
      fi
    fi
    if [ $AGENT_INST -eq 1 ]; then
      echo "  - Fabasoft app.telemetry Agent"
      PACKAGES="$PACKAGES Agent/$RPMPATH/apptelemetryagent*"
    fi
    if [ $WEBAPI_INST -eq 1 ]; then
      echo "  - Fabasoft app.telemetry WebAPI"
      PACKAGES="$PACKAGES TelemetryWeb/$RPMPATH/apptelemetryweb*"
    fi
    read -n 1 -p "Do you want to upgrade Fabasoft app.telemetry? [Y] " UPG

    if [ -z "$UPG" ]; then
      UPG="Y"
    else
      echo
    fi

    if [ "$UPG" = "Y" -o "$UPG" = "y" ]; then
      UPG="Y"
      break
    elif [ "$UPG" = "N" -o "$UPG" = "n" ]; then
      UPG="N"
        echo "Upgrade canceled."
        exit 0
      break
    fi
  done

  RPMCMD="rpm -Uhv $PACKAGES"
fi

# --------------------------------------------------------------------------
#  New installation
# --------------------------------------------------------------------------

if [ "$INST_TYPE" = "INSTALL" ] ; then
  while true; do
    echo
    echo "Specify the Fabasoft app.telemetry packages to install:"
    if [ "$OS" = "EL5" ]; then
      echo "  !  Fabasoft app.telemetry Server is not supported any longer for this platform! (you should update to RHEL/CentOS 7.x)"
    else
      echo "  A) Fabasoft app.telemetry Server"
    fi
    echo "  B) Fabasoft app.telemetry Agent"
    echo "  C) Fabasoft app.telemetry Agent with WebAPI (end-user experience monitoring)"
    read -n 1 -p "Which Fabasoft app.telemetry packages do you want to install? [$DEF] " PKG

    if [ -z "$PKG" ]; then
      PKG="$DEF"
    else
      echo
    fi

    if [ "$PKG" = "A" -o "$PKG" = "a" ]; then
      if [ "$OS" = "EL5" ]; then
        echo "  WARNING: Fabasoft app.telemetry Server is not supported any longer for this platform! (please update to RHEL/CentOS 7.x)"
      else
        RPMCMD="rpm -ihv Server/$RPMPATH/apptelemetryserver* Agent/$RPMPATH/apptelemetryagent* TelemetryWeb/$RPMPATH/apptelemetryweb*"
        SERVERSETUP=yes
      fi
      break
    elif [ "$PKG" = "B" -o "$PKG" = "b" ]; then
      RPMCMD="rpm -ihv Agent/$RPMPATH/apptelemetryagent*"
      break
    elif [ "$PKG" = "C" -o "$PKG" = "c" ]; then
      RPMCMD="rpm -ihv Agent/$RPMPATH/apptelemetryagent* TelemetryWeb/$RPMPATH/apptelemetryweb*"
      break
    fi
  done
fi

# --------------------------------------------------------------------------
#  RPM installation/upgrade und serversetup.sh
# --------------------------------------------------------------------------
echo
(
  cd "$(dirname "$0")"
  $RPMCMD 2>&1 | fgrep -v "the first time"
  RPM_EXIT_CODE=${PIPESTATUS[0]}
  GREP_EXIT_CODE=$?

  if [ $RPM_EXIT_CODE -eq 0 -a $GREP_EXIT_CODE -eq 0 -a "$SERVERSETUP" = "yes" -a -f /opt/app.telemetry/bin/serversetup.sh ]; then
    /opt/app.telemetry/bin/serversetup.sh -s
  fi

  if [ "$OS" != "EL7" -a -d /tmp/app.telemetry ]; then    # fix possible wrong filesystem permissions for webserver
   chmod 777 /tmp/app.telemetry
  fi
)
