#!/bin/bash
# ===========================================================================
# Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1995-2015.
#
# All rights reserved. All hardware and software names used are registered
# trade names and/or registered trademarks of the respective manufacturers.
#
# The user of this computer program acknowledges that the above copyright
# notice, which constitutes the Universal Copyright Convention, will be
# attached at the position in the function of the computer program which the
# author has deemed to sufficiently express the reservation of copyright.
# It is prohibited for customers, users and/or third parties to remove,
# modify or move this copyright notice.
# ==========================================================================

if [ "$1" != "-s" ]; then
  echo
  echo "  Fabasoft app.telemetry Server Setup"
  echo
  echo "  Copyright (c) Fabasoft R&D GmbH, A-4020 Linz, 1988-2015."
  echo
fi

# --------------------------------------------------------------------------
#  Initialize
# --------------------------------------------------------------------------
LOG_FILE=/var/log/app.telemetry/serversetup.log
LICENSE_FILE=/etc/app.telemetry/license.lic
APM_HTPASSWD_FILE=/etc/app.telemetry/htpasswd
APM_HTGROUP_FILE=/etc/app.telemetry/htgroup
APM_ADMIN_HTGROUP=apptelemetryadministrators
APM_USER_HTGROUP=apptelemetryusers
APM_DASHBOARD_HTGROUP=apptelemetrydashboardusers
APM_LOGPOOL_HTGROUP=apptelemetrylogpoolusers
APM_WEBFORM_HTGROUP=apptelemetrywebformusers
RHEL_OS="RHEL/CentOS"
DEFAULT_ADMIN_USER="admin"
declare -i W=0

DEFLATE_TEMPLATE=" \t<IfModule mod_deflate.c>\n"
DEFLATE_TEMPLATE="$DEFLATE_TEMPLATE \t   SetOutputFilter DEFLATE\n"
DEFLATE_TEMPLATE="$DEFLATE_TEMPLATE \t</IfModule>\n"

OS=$RHEL_OS
USED_HTPASSWD_CMD=htpasswd
USED_HTTPD_DAEMON_NAME=httpd
USED_APACHE_CONF_FILE=/etc/httpd/conf.d/apptelemetrywebserver.conf
APACHE_USER=apache
APACHE_GRP=apache

INST_VER=`rpm -q --queryformat '%{VERSION}' apptelemetryserver`

MAJOR_VERSION=0
if [ -f /etc/os-release ] ; then
  . /etc/os-release
  if [ "$ID" = "centos" -o "$ID" = "rhel" ] ; then
    MAJOR_VERSION=$(echo "$VERSION_ID" | cut -d "." -f -1)
  fi
fi

echo "" >>$LOG_FILE
echo "Fabasoft app.telemetry Server Setup (on: $OS, date: `date`, version: $INST_VER)  " >>$LOG_FILE
echo "-----------------------------------  " >>$LOG_FILE

# --------------------------------------------------------------------------
#  Backup old config files
# --------------------------------------------------------------------------
echo " § Backup configuration files:" >>$LOG_FILE 2>&1
if [[ -f $APM_HTPASSWD_FILE ]] ; then
  cp -fv $APM_HTPASSWD_FILE ${APM_HTPASSWD_FILE}.bak >>$LOG_FILE 2>&1
fi
if [[ -f $APM_HTGROUP_FILE ]] ; then
  cp -f $APM_HTGROUP_FILE ${APM_HTGROUP_FILE}.bak >>$LOG_FILE 2>&1
fi
if [[ -f $USED_APACHE_CONF_FILE ]] ; then
  cp -f $USED_APACHE_CONF_FILE ${USED_APACHE_CONF_FILE}.bak >>$LOG_FILE 2>&1
fi

# --------------------------------------------------------------------------
#  New installation
# --------------------------------------------------------------------------
if test ! -f $LICENSE_FILE ; then
    while true; do
      echo
      read -e -p "Specify the Fabasoft app.telemetry license file (.lic) to load: " USER_LICENSE_FILE

      if [ -f "$USER_LICENSE_FILE" ] ; then
        cp -f "$USER_LICENSE_FILE" $LICENSE_FILE
        echo " -> installed new license file to $LICENSE_FILE."
        break
      else
        echo "File not found."
      fi
    done
fi

# --------------------------------------------------------------------------
#  Apache authentication settings
# --------------------------------------------------------------------------
echo "" >>$LOG_FILE 2>&1
echo " § Check and update/upgrade Apache authentication settings:" >>$LOG_FILE 2>&1
if [ ! -f $APM_HTPASSWD_FILE ] ; then
  echo
  read -p "Specify the Fabasoft app.telemetry administrator [$DEFAULT_ADMIN_USER]: " ADMIN_USER
  if [[ -z $ADMIN_USER ]] ; then
    ADMIN_USER=$DEFAULT_ADMIN_USER
  fi

  echo "Provide a password for the Fabasoft app.telemetry administrator"
  while true; do
    if $USED_HTPASSWD_CMD -c $APM_HTPASSWD_FILE $ADMIN_USER ; then
      break;
    fi
  done
  chmod 640 $APM_HTPASSWD_FILE
  chown $APACHE_USER:$APACHE_GRP $APM_HTPASSWD_FILE
fi

echo

(
  if [[ -f $APM_HTGROUP_FILE ]] ; then
    echo "  Checking Apache authentication group settings for app.telemetry: "
    echo "    Basic authentication group file '$APM_HTGROUP_FILE' already exists. Checking existing authentication groups ..."
    if grep -i "apptelemetryadministrators:" $APM_HTGROUP_FILE | grep -i -v $ADMIN_USER >/dev/null ; then
      echo "      Selected admin user is not in administrators group -> adding it ..."
      sed -i -s -e s/apptelemetryadministrators:/apptelemetryadministrators:\ $ADMIN_USER\ / $APM_HTGROUP_FILE
    fi
    if ! grep -i "apptelemetrydashboardusers:" $APM_HTGROUP_FILE ; then
      echo "      No apptelemetrydashboardusers group found -> adding it ..."
      echo "$APM_DASHBOARD_HTGROUP: " >> $APM_HTGROUP_FILE
    fi
    if ! grep -i "apptelemetrylogpoolusers:" $APM_HTGROUP_FILE ; then
      echo "      No apptelemetrylogpoolusers group found -> adding it ..."
      echo "$APM_LOGPOOL_HTGROUP: " >> $APM_HTGROUP_FILE
    fi
    if ! grep -i "apptelemetrywebformusers:" $APM_HTGROUP_FILE ; then
      echo "      No apptelemetrywebformusers group found -> adding it ..."
      echo "$APM_WEBFORM_HTGROUP: " >> $APM_HTGROUP_FILE
    fi
    chmod 640 $APM_HTGROUP_FILE
    chown $APACHE_USER:$APACHE_GRP $APM_HTGROUP_FILE
    echo "    -> new file content ..."
    cat $APM_HTGROUP_FILE

  else
    echo "$APM_ADMIN_HTGROUP: $ADMIN_USER" > $APM_HTGROUP_FILE
    echo "$APM_USER_HTGROUP: " >> $APM_HTGROUP_FILE
    echo "$APM_DASHBOARD_HTGROUP: " >> $APM_HTGROUP_FILE
    echo "$APM_LOGPOOL_HTGROUP: " >> $APM_HTGROUP_FILE
    echo "$APM_WEBFORM_HTGROUP: " >> $APM_HTGROUP_FILE
    chmod 640 $APM_HTGROUP_FILE
    chown $APACHE_USER:$APACHE_GRP $APM_HTGROUP_FILE
    echo "   -> Created basic authentication group '$APM_ADMIN_HTGROUP' (including user '$ADMIN_USER'): $APM_HTGROUP_FILE"
  fi
  echo

  if [[ -f $USED_APACHE_CONF_FILE ]] ; then
    echo "  Checking Apache module handler settings for app.telemetry: "
    if ! grep -i -C 3 "SetOutputFilter DEFLATE" $USED_APACHE_CONF_FILE ; then
      echo "    WARNING: no DEFLATE outputfilter detected in file '$USED_APACHE_CONF_FILE'."
      echo "    -> You should manually add content compression to the base app.telemetry web directory directive ..."
      echo "        $DEFLATE_TEMPLATE"
      let "W++"
    fi
  fi
  echo
  echo
) >> $LOG_FILE 2>&1


# --------------------------------------------------------------------------
#  Daemon settings
# --------------------------------------------------------------------------
echo " § Check and restart daemons:" >>$LOG_FILE 2>&1

if [ "$MAJOR_VERSION" = "7" ] ; then
  /usr/bin/systemctl enable httpd.service | tee -a $LOG_FILE
  /usr/bin/systemctl reload-or-restart httpd.service | tee -a $LOG_FILE
  if [ -f /usr/lib/systemd/system/snmpd.service ] ; then
    /usr/bin/systemctl enable snmpd.service | tee -a $LOG_FILE
    /usr/bin/systemctl reload-or-restart snmpd.service | tee -a $LOG_FILE
  else
    echo "  WARNING: snmp daemon (unit /usr/lib/systemd/system/snmpd.service) not found!" >> $LOG_FILE
    echo "           It is recommended to install the net-snmp RPM package." >> $LOG_FILE
    let "W++"
  fi
  /usr/bin/systemctl restart apptelemetryserver.service | tee -a $LOG_FILE
  /usr/bin/systemctl restart apptelemetryecomm.service | tee -a $LOG_FILE
  /usr/bin/systemctl restart apptelemetryworker.service | tee -a $LOG_FILE
  /usr/bin/systemctl restart apptelemetryagent.service | tee -a $LOG_FILE
else
  chkconfig --level 345 $USED_HTTPD_DAEMON_NAME on | tee -a $LOG_FILE
  /etc/init.d/$USED_HTTPD_DAEMON_NAME status >/dev/null 2>&1
  if [ "$?" -eq 0 ]; then
    /etc/init.d/$USED_HTTPD_DAEMON_NAME stop | tee -a $LOG_FILE
  fi
  /etc/init.d/$USED_HTTPD_DAEMON_NAME start | tee -a $LOG_FILE
  if [[ -f "/etc/init.d/snmpd" ]] ; then
    chkconfig --level 345 snmpd on | tee -a $LOG_FILE
    /etc/init.d/snmpd status >/dev/null 2>&1
    if [ "$?" -eq 0 ]; then
      /etc/init.d/snmpd stop | tee -a $LOG_FILE
    fi
    /etc/init.d/snmpd start | tee -a $LOG_FILE
  else
    echo "  WARNING: no snmp daemon (init script /etc/init.d/snmpd) found!" >> $LOG_FILE
    echo "           It is recommended to install the net-snmp RPM package." >> $LOG_FILE
    let "W++"
  fi
  /etc/init.d/apptelemetryserverd status >/dev/null 2>&1
  if [ "$?" -eq 0 ]; then
    /etc/init.d/apptelemetryserverd stop | tee -a $LOG_FILE
  fi
  /etc/init.d/apptelemetryserverd start | tee -a $LOG_FILE
  /etc/init.d/apptelemetryagentd status >/dev/null 2>&1
  if [ "$?" -eq 0 ]; then
    /etc/init.d/apptelemetryagentd stop | tee -a $LOG_FILE
  fi
  /etc/init.d/apptelemetryagentd start | tee -a $LOG_FILE
fi

if [ $W -gt 0 ]; then
  echo "WARNING: $W warnings occured during setup! For details see logfile: $LOG_FILE"
fi

echo
echo "**********************************************************************"
echo "*                                                                    *"
echo "* Fabasoft app.telemetry configuration completed. You can connect    *"
echo "* with your web browser using the following URL:                     *"
echo "*                                                                    *"
echo "            http://$(hostname)/apptelemetry/"
echo "*                                                                    *"
echo "*         For more details visit www.apptelemetry.com                *"
echo "*                                                                    *"
echo "**********************************************************************"
echo

echo "" >> $LOG_FILE
echo "Fabasoft app.telemetry Setup COMPLETE." >> $LOG_FILE
echo "========================================" >> $LOG_FILE

